from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from ultrack.core.export.trackmate import tracks_layer_to_trackmate
from ultrack.utils.constants import NO_PARENT

pytrackmate = pytest.importorskip("pytrackmate")


def test_trackmate_export_spot_match(tmp_path: Path) -> None:
    """Check if the spots (objects) match between the tracks and the exported trackmate xml file.

    This test cannot check if the exported tracking is valid.
    """
    tracks_outpath = tmp_path / "tracks.xml"

    tracks_df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4],
            "parent_id": [NO_PARENT, 1, 2, 2],
            "track_id": [1, 1, 2, 3],
            "t": [0, 1, 2, 2],
            "z": [0, 0, 0, 0],
            "y": [10, 20, 30, 10],
            "x": [1, 2, 3, 1],
        }
    )

    xml_str = tracks_layer_to_trackmate(tracks_df)
    with open(tracks_outpath, "w") as f:
        f.write(xml_str)

    trackmate_df = pytrackmate.trackmate_peak_import(tracks_outpath)
    print(trackmate_df)

    assert trackmate_df.shape[0] == tracks_df.shape[0]

    np.testing.assert_allclose(
        tracks_df[["t", "z", "y", "x"]], trackmate_df[["t_stamp", "z", "y", "x"]]
    )



from magicgui.widgets import ComboBox, FloatSpinBox, PushButton, SpinBox

from ultrack.config.segmentationconfig import NAME_TO_WS_HIER, SegmentationConfig
from ultrack.widgets.ultrackwidget._legacy.baseconfigwidget import BaseConfigWidget


class SegmentationWidget(BaseConfigWidget):
    def __init__(self, config: SegmentationConfig):
        super().__init__(label="Segmentation", config=config)

    def _setup_widgets(self) -> None:
        self._attr_to_widget = {
            "threshold": FloatSpinBox(label="Det. threshold"),
            "min_area": SpinBox(label="Min. area", max=1_000_000_000),
            "max_area": SpinBox(label="Max. area", max=1_000_000_000),
            "min_frontier": FloatSpinBox(label="Min. frontier"),
            "anisotropy_penalization": FloatSpinBox(
                label="Anisotropy pen.", min=-100, max=100
            ),
            "ws_hierarchy": ComboBox(
                label="Watershed by", choices=list(NAME_TO_WS_HIER.items())
            ),
            "n_workers": SpinBox(label="Num. workers"),
        }
        for widget in self._attr_to_widget.values():
            self.append(widget)

        self._segment_btn = PushButton(text="Segment")
        self.append(self._segment_btn)



import json
from pathlib import Path
from typing import Any, Callable, Dict, Optional

from napari.layers import Image, Layer
from qtpy.QtWidgets import (
    QCheckBox,
    QGroupBox,
    QLabel,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from ultrack.config.config import CFG_ALIAS_TO_ATTR, MainConfig
from ultrack.widgets.ultrackwidget.components.blankable_number_edit import (
    BlankableNumberEdit,
)
from ultrack.widgets.ultrackwidget.components.no_wheel_spin import (
    NoWheelDoubleSpinBox,
    NoWheelSpinBox,
)
from ultrack.widgets.ultrackwidget.workflows import WorkflowChoice


class DataForms:
    """
    A class to manage and display data forms for the Ultrack configuration.

    Attributes
    ----------
    _tab : QTabWidget
        The main tab widget containing all forms.
    _bindings : list
        Bindings for standard form fields.
    _additional_bindings : list
        Bindings for additional form fields.
    _update_channel_axis_bindings : list
        Bindings for updating the channel axis field based on image data.
    _additional_forms : dict
        Additional forms for optional configuration sections.

    Methods
    -------
    _setup_forms()
        Setup and initialize the forms based on metadata.
    load_config(config: MainConfig)
        Load the configuration into the form fields.
    _create_form(id_form: str, metadata: Dict[str, Any])
        Create a form based on provided metadata.
    _create_field(layout: QVBoxLayout, id_form: str, id_field: str, field_metadata: Dict, current_subform: str)
        Create an individual form field.
    _create_checkbox_field(layout: QVBoxLayout, id_form: str, id_field: str, field_metadata: Dict, current_subform: str)
        Create a checkbox field.
    _create_number_field(layout: QVBoxLayout, id_form: str, id_field: str, field_metadata: Dict, current_subform: str)
        Create a number input field.
    get_additional_options() -> Dict[str, Any]
        Get the additional options from the form.
    get_config() -> MainConfig
        Get the current configuration from the form.
    setup_additional_options(additional_options: List[str])
        Setup visibility of additional options based on input.
    get_tab_widget() -> QTabWidget
        Get the main tab widget.
    """

    def __init__(self, on_change_any_input: Callable) -> None:
        """
        Initialize the DataForms class.

        Parameters
        ----------
        on_change_any_input : Callable
            Callback function to be called when any input changes.
        """
        self.on_change_any_input = on_change_any_input

        self._config = None
        self._tab = QTabWidget()
        self._bindings = []
        self._update_channel_axis_bindings: list[BlankableNumberEdit] = []
        self._additional_bindings = []
        self._additional_forms = {}
        self._setup_forms()
        self.load_config(MainConfig())

    def _setup_forms(self) -> None:
        """
        Setup and initialize the forms based on metadata.
        """
        rsrc_dir = Path(__file__).parent / "resources"
        with (
            open(rsrc_dir / "forms.json") as forms,
            open(rsrc_dir / "additional_options.json") as opts,
        ):
            forms_metadata = json.load(forms)
            additional_options = json.load(opts)
            forms_metadata["additional_options"] = {
                "title": "Pre-processing",
                "fields": additional_options,
            }

        self._create_form("additional_options", forms_metadata["additional_options"])
        self._create_form("segmentation", forms_metadata["segmentation"])
        self._create_form("linking", forms_metadata["linking"])
        self._create_form("tracking", forms_metadata["tracking"])

    def load_config(self, config: MainConfig) -> None:
        """
        Load the configuration into the form fields.

        Parameters
        ----------
        config : MainConfig
            The main configuration to load.
        """
        self._config = config
        for id_form, id_field, widget, _, setter in self._bindings:
            # Get the actual config field name from the alias mapping
            field_name = CFG_ALIAS_TO_ATTR.get(id_form, id_form)
            # In Pydantic v2, we can still use getattr for nested access
            value = getattr(getattr(self._config, field_name), id_field)
            getattr(widget, setter)(value)

    def _create_form(self, id_form: str, metadata: Dict[str, Any]) -> None:
        """
        Create a form based on provided metadata.

        Parameters
        ----------
        id_form : str
            The form identifier.
        metadata : Dict[str, Any]
            Metadata describing the form fields and layout.
        """
        title = metadata["title"]
        fields = metadata["fields"]
        tab = QWidget()
        layout = QVBoxLayout()
        main_layout = layout
        tab.setLayout(layout)

        fields_to_be_added = [*fields.items()]
        current_subform = None
        while fields_to_be_added:
            id_field, field_metadata = fields_to_be_added.pop(0)

            if not field_metadata.get("subform", False):
                self._create_field(
                    layout, id_form, id_field, field_metadata, current_subform
                )
            else:
                subgroup = QGroupBox(field_metadata["title"])
                current_subform = id_field
                layout = QVBoxLayout()
                subgroup.setLayout(layout)
                fields_to_be_added = [
                    *field_metadata["fields"].items()
                ] + fields_to_be_added
                main_layout.addWidget(subgroup)
                self._additional_forms[id_field] = subgroup

        main_layout.addStretch()
        self._tab.addTab(tab, title)

    def _create_field(
        self,
        layout: QVBoxLayout,
        id_form: str,
        id_field: str,
        field_metadata: Dict,
        current_subform: Optional[str],
    ) -> None:
        """
        Create an individual form field.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the field will be added.
        id_form : str
            The form identifier.
        id_field : str
            The field identifier.
        field_metadata : Dict
            Metadata describing the field.
        current_subform : Optional[str]
            The current subform identifier if within a subform.
        """
        if field_metadata["type"] == "number":
            label = QLabel(f'{field_metadata["label"]} {chr(0xF059)}')
            label.setToolTip(field_metadata["tooltip"])
            layout.addWidget(label)

            self._create_number_field(
                layout, id_form, id_field, field_metadata, current_subform
            )
        elif field_metadata["type"] == "checkbox":
            self._create_checkbox_field(
                layout, id_form, id_field, field_metadata, current_subform
            )
        else:
            raise NotImplementedError(
                f"Field type {field_metadata['type']} not implemented"
            )

    def _create_checkbox_field(
        self,
        layout: QVBoxLayout,
        id_form: str,
        id_field: str,
        field_metadata: Dict,
        current_subform: Optional[str],
    ) -> None:
        """
        Create a checkbox field.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the field will be added.
        id_form : str
            The form identifier.
        id_field : str
            The field identifier.
        field_metadata : Dict
            Metadata describing the field.
        current_subform : Optional[str]
            The current subform identifier if within a subform.
        """
        checkbox = QCheckBox(f'{field_metadata["label"]} {chr(0xF059)}')
        checkbox.setToolTip(field_metadata["tooltip"])
        checkbox.setChecked(bool(field_metadata.get("default", False)))
        checkbox.stateChanged.connect(self.on_change_any_input)
        layout.addWidget(checkbox)
        if id_form != "additional_options":
            self._bindings.append(
                (id_form, id_field, checkbox, "isChecked", "setChecked")
            )
        else:
            self._additional_bindings.append(
                (current_subform, id_field, checkbox, "isChecked")
            )

        if id_field == "__enable__":
            checkbox.stateChanged.connect(
                lambda state, subform=current_subform, _checkbox=checkbox: self._disable_subform(
                    state, self._additional_forms[subform], _checkbox
                )
            )

    def _disable_subform(
        self, state: bool, subform: QGroupBox, caller: QCheckBox
    ) -> None:
        """Disable all widgets in a subform except the caller.

        Parameters
        ----------
        state : bool
            The state to set the widgets to.
        subform : QGroupBox
            The subform to disable.
        caller : QCheckBox
            The caller widget that should remain enabled.
        """
        for widget in subform.children():
            if widget != caller:
                widget.setEnabled(state)

    def _create_number_field(
        self,
        layout: QVBoxLayout,
        id_form: str,
        id_field: str,
        field_metadata: Dict,
        current_subform: Optional[str],
    ) -> None:
        """
        Create a number input field.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the field will be added.
        id_form : str
            The form identifier.
        id_field : str
            The field identifier.
        field_metadata : Dict
            Metadata describing the field.
        current_subform : Optional[str]
            The current subform identifier if within a subform.
        """
        dtype = float if field_metadata.get("step") == "any" else int
        max_value = dtype(field_metadata.get("max", 999999999))
        min_value = dtype(field_metadata.get("min", -999999999))
        default = field_metadata.get("default", 0)

        if field_metadata.get("required", False):
            spin_box = NoWheelDoubleSpinBox() if dtype == float else NoWheelSpinBox()
            if dtype == float:
                spin_box.setDecimals(5)
            spin_box.setMinimum(min_value)
            spin_box.setMaximum(max_value)
            spin_box.setValue(dtype(default))
            layout.addWidget(spin_box)
            spin_box.valueChanged.connect(self.on_change_any_input)

            if id_form != "additional_options":
                self._bindings.append(
                    (id_form, id_field, spin_box, "value", "setValue")
                )
            else:
                self._additional_bindings.append(
                    (current_subform, id_field, spin_box, "value")
                )
        else:
            number_edit = BlankableNumberEdit(
                default=default,
                dtype=dtype,
                minimum=min_value,
                maximum=max_value,
            )
            layout.addWidget(number_edit)
            number_edit.textChanged.connect(self.on_change_any_input)

            if id_field == "channel_axis":
                self._update_channel_axis_bindings.append(number_edit)

            if id_form != "additional_options":
                self._bindings.append(
                    (
                        id_form,
                        id_field,
                        number_edit,
                        "getValue",
                        "setText",
                    )
                )
            else:
                self._additional_bindings.append(
                    (
                        current_subform,
                        id_field,
                        number_edit,
                        "getValue",
                    )
                )

    def get_additional_options(self) -> Dict[str, Any]:
        """
        Get the additional options from the form.

        Returns
        -------
        Dict[str, Any]
            A dictionary of additional options.
        """
        additional_options = {}
        for id_form, id_field, widget, getter in self._additional_bindings:
            if not self._additional_forms[id_form].isHidden():
                value = getattr(widget, getter)()
                if id_form not in additional_options:
                    additional_options[id_form] = {}
                additional_options[id_form][id_field] = value
        return additional_options

    def get_config(self) -> MainConfig:
        """
        Get the current configuration from the form.

        Returns
        -------
        MainConfig
            The main configuration object.
        """
        for id_form, id_field, widget, getter, _ in self._bindings:
            value = getattr(widget, getter)()
            # Get the actual config field name from the alias mapping
            field_name = CFG_ALIAS_TO_ATTR.get(id_form, id_form)
            sub_config = getattr(self._config, field_name)
            # In Pydantic v2, we can still use setattr for direct updates
            setattr(sub_config, id_field, value)
        return self._config

    def setup_additional_options(self, workflow_choice: WorkflowChoice) -> None:
        """
        Setup visibility of additional options based on input.

        Each workflow choice has a different set of additional options. This method
        will show or hide the additional options based on the workflow choice.
        The additional options are contained in the field `_additional_forms` and
        maps the form id to the form widget. Since the workflow maps to the form id,
        we can show or hide the form based on the workflow choice.

        Parameters
        ----------
        workflow_choice : str
            The workflow choice to determine which additional options to show.
        """
        map_workflow_to_form_ids = {
            WorkflowChoice.AUTO_FROM_LABELS: ["label_to_contours_kwargs"],
            WorkflowChoice.AUTO_DETECT: [
                "detect_foreground_kwargs",
                "robust_invert_kwargs",
                "flow_kwargs",
            ],
            WorkflowChoice.MANUAL: ["flow_kwargs"],
        }

        any_visible = False
        for id_options in self._additional_forms.keys():
            if id_options in map_workflow_to_form_ids[workflow_choice]:
                any_visible = True
                self._additional_forms[id_options].show()
            else:
                self._additional_forms[id_options].hide()

        # hard-coded index for additional options tab
        # if no additional options are visible, hide the tab
        self._tab.setTabVisible(0, any_visible)

    def notify_image_update(self, image: Layer) -> None:
        channel = ""
        if image and isinstance(image, Image) and image.rgb:
            channel = image.data.ndim - 2  # last dimension is channel

        for widget in self._update_channel_axis_bindings:
            widget.setText(str(channel))

    def get_tab_widget(self) -> QTabWidget:
        """
        Get the main tab widget.

        Returns
        -------
        QTabWidget
            The main tab widget containing all forms.
        """
        return self._tab



import logging
from typing import Dict, List, Optional, Set, Tuple, Union

import numpy as np
import pandas as pd
from numba import njit, typed, types
from numpy.typing import ArrayLike
from zarr.storage import StoreLike

from ultrack.utils.constants import NO_PARENT
from ultrack.utils.segmentation import SegmentationPainter, copy_segments

LOG = logging.getLogger(__name__)


@njit
def _fast_path_transverse(
    node: int,
    track_id: int,
    queue: List[Tuple[int, int]],
    forest: Dict[int, Tuple[int]],
) -> List[int]:
    """Transverse a path in the forest directed graph and add path (track) split into queue.

    Parameters
    ----------
    node : int
        Source path node.
    track_id : int
        Reference track id for path split.
    queue : List[Tuple[int, int]]
        Source nodes and path (track) id reference queue.
    forest : Dict[int, Tuple[int]]
        Directed graph (tree) of paths relationships.

    Returns
    -------
    List[int]
        Sequence of nodes in the path.
    """
    path = typed.List.empty_list(types.int64)

    while True:
        path.append(node)

        children = forest.get(node)
        if children is None:
            # end of track
            break

        elif len(children) == 1:
            node = children[0]

        elif len(children) == 2:
            queue.append((children[1], track_id))
            queue.append((children[0], track_id))
            break

        else:
            raise RuntimeError(
                "Something is wrong. Found node with more than two children when parsing tracks."
            )

    return path


@njit
def _fast_forest_transverse(
    roots: List[int],
    forest: Dict[int, List[int]],
) -> Tuple[List[List[int]], List[int], List[int], List[int]]:
    """Transverse the tracks forest graph creating a distinc id to each path.

    Parameters
    ----------
    roots : List[int]
        Forest roots.
    forest : Dict[int, List[int]]
        Graph (forest).

    Returns
    -------
    Tuple[List[List[int]], List[int], List[int], List[int]]
        Sequence of paths, their respective track_id, parent_track_id and length.
    """
    track_id = 1
    paths = []
    track_ids = []  # equivalent to arange
    parent_track_ids = []
    lengths = []

    for root in roots:
        queue = [(root, NO_PARENT)]

        while queue:
            node, parent_track_id = queue.pop()
            path = _fast_path_transverse(node, track_id, queue, forest)
            paths.append(path)
            track_ids.append(track_id)
            parent_track_ids.append(parent_track_id)
            lengths.append(len(path))
            track_id += 1

    return paths, track_ids, parent_track_ids, lengths


@njit
def _create_tracks_forest(
    node_ids: np.ndarray, parent_ids: np.ndarray
) -> Dict[int, List[int]]:
    """Creates the forest graph of track lineages

    Parameters
    ----------
    node_ids : np.ndarray
        Nodes indices.
    parent_ids : np.ndarray
        Parent indices.

    Returns
    -------
    Dict[int, List[int]]
        Forest graph where parent maps to their children (parent -> children)
    """
    forest = {}
    for parent in parent_ids:
        forest[parent] = typed.List.empty_list(types.int64)

    for i in range(len(parent_ids)):
        forest[parent_ids[i]].append(node_ids[i])

    return forest


def add_track_ids_to_tracks_df(df: pd.DataFrame) -> pd.DataFrame:
    """Adds `track_id` and `parent_track_id` columns to forest `df`.
    Each maximal path receveis a unique `track_id`.

    Parameters
    ----------
    df : pd.DataFrame
        Forest defined by the `parent_id` column and the dataframe indices.

    Returns
    -------
    pd.DataFrame
        Inplace modified input dataframe with additional columns.
    """
    assert df.shape[0] > 0

    df.index = df.index.astype(int)
    df["parent_id"] = df["parent_id"].astype(int)

    forest = _create_tracks_forest(df.index.values, df["parent_id"].values)
    roots = forest.pop(NO_PARENT)

    df["track_id"] = NO_PARENT
    df["parent_track_id"] = NO_PARENT

    paths, track_ids, parent_track_ids, lengths = _fast_forest_transverse(roots, forest)

    paths = np.concatenate(paths)
    df.loc[paths, "track_id"] = np.repeat(track_ids, lengths)
    df.loc[paths, "parent_track_id"] = np.repeat(parent_track_ids, lengths)

    unlabeled_tracks = df["track_id"] == NO_PARENT
    assert not np.any(
        unlabeled_tracks
    ), f"Something went wrong. Found unlabeled tracks\n{df[unlabeled_tracks]}"

    return df


def tracks_df_forest(
    df: pd.DataFrame,
    remove_roots: bool = False,
    numba_dict: bool = False,
) -> Union[
    Dict[int, List[int]], types.DictType(types.int64, types.ListType(types.int64))
]:
    """
    Creates the forest graph of track lineages

    Example:
    forest[parent_id] = [child_id_0, child_id_1]

    Parameters
    ----------
    df : pd.DataFrame
        Tracks dataframe.
    remove_roots : bool
        If True, removes root nodes (nodes with no parent).
    numba_dict: bool
        If True, returns a numba typed dictionary.

    Returns
    -------
    Dict[int, List[int]]
        Forest graph where parent maps to their children (parent -> children)
    """
    df = df.drop_duplicates("track_id")

    if remove_roots:
        df = df[df["parent_track_id"] != NO_PARENT]

    nb_dict = _create_tracks_forest(
        df["track_id"].to_numpy(dtype=int),
        df["parent_track_id"].to_numpy(dtype=int),
    )

    if numba_dict:
        return nb_dict

    return {k: v for k, v in nb_dict.items()}


def inv_tracks_df_forest(df: pd.DataFrame) -> Dict[int, int]:
    """
    Returns `track_id` and `parent_track_id` leaves-to-root inverted forest (set of trees) graph structure.

    Example:
    forest[child_id] = parent_id
    """
    for col in ["track_id", "parent_track_id"]:
        if col not in df.columns:
            raise ValueError(
                f"The input dataframe does not contain the column '{col}'."
            )

    df = df.drop_duplicates("track_id")
    df = df[df["parent_track_id"] != NO_PARENT]
    graph = {}
    for parent_id, id in zip(df["parent_track_id"], df["track_id"]):
        graph[id] = parent_id
    return graph


@njit
def left_first_search(
    track_id: int,
    graph: types.DictType(types.int64, types.ListType(types.int64)),
) -> List[int]:
    """
    Perform a left-first traversal on a binary tree represented as a graph and return a list
    of track IDs in the order they are visited during the traversal.

    Parameters
    ----------
    track_id : int
        The ID of the track to start the traversal from.
    graph : Dict[int, List[int]]
        The graph representing the binary tree. It is a dictionary where the keys are track IDs,
        and the values are lists of two child track IDs. The binary tree must have exactly two
        children for each node.

    Returns
    -------
    List[int]
        A list of track IDs visited during the left-first traversal of the binary tree,
        with `track_id` being the starting point.

    Example
    -------
    >>> graph = {1: [2, 3], 2: [4, 5], 3: [6, 7], 4: None, 5: None, 6: None, 7: None}
    >>> result = _left_first_search(1, graph)
    >>> print(result)
    [4, 2, 5, 1, 6, 3, 7]
    """

    children = graph.get(track_id)

    if children is None:
        return [track_id]

    assert len(children) == 2

    sorted_track_ids = left_first_search(children[0], graph)
    sorted_track_ids.append(track_id)
    sorted_track_ids += left_first_search(children[1], graph)

    return sorted_track_ids


def get_subgraph(
    tracks_df: pd.DataFrame,
    track_ids: ArrayLike,
) -> pd.DataFrame:
    """
    Get a subgraph from a forest of tracks represented as a DataFrame.

    Parameters
    ----------
    tracks_df : pd.DataFrame
        DataFrame containing track information with columns:
            "track_id" : Unique identifier for each track.
            "parent_track_id" : Identifier of the parent track in the forest.
            (Other columns may be present in the DataFrame but are not used in this function.)
    track_ids : ArrayLike
        An array-like object containing the track IDs for which to extract the subgraph.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the subgraph of tracks corresponding to the input track IDs.

    Examples
    --------
    >>> subgraph_df = get_subgraph(tracks_df, [3, 7, 10])

    Notes
    -----
    The input DataFrame 'tracks_df' should have at least two columns: "track_id" and "parent_track_id",
    where "track_id" represents the unique identifier for each track, and "parent_track_id" represents
    the identifier of the parent track in the forest.
    """
    track_ids = np.atleast_1d(track_ids).astype(int)
    compressed_df = tracks_df.drop_duplicates("track_id")

    inv_graph = inv_tracks_df_forest(compressed_df)
    selected_roots = set()
    for id in track_ids:

        while True:
            parent_id = inv_graph.get(id, NO_PARENT)
            if parent_id == NO_PARENT:
                break
            id = parent_id

        selected_roots.add(id)

    graph = _create_tracks_forest(
        compressed_df["track_id"].to_numpy(dtype=int),
        compressed_df["parent_track_id"].to_numpy(dtype=int),
    )
    selected_roots = np.asarray(list(selected_roots), dtype=int)

    subforest = []
    for root in selected_roots:
        subforest += left_first_search(root, graph)

    return tracks_df[tracks_df["track_id"].isin(subforest)]


def get_subtree(graph: Dict[int, int], index: int) -> Set[int]:
    """
    Returns connected component of directed graph (subtree) of `index` of tree `graph`.

    Parameters
    ----------
    graph : Dict[int, int]
        Directed graph (forest) from Parent -> Child.
    index : int
        Index of the root node.

    Returns
    -------
    Set[int]
        Set of nodes in the subtree.
    """
    component = set()
    queue = [index]
    while queue:
        index = queue.pop()
        component.add(index)
        for child in graph.get(index, []):
            queue.append(child)

    return component


def split_trees(tracks_df: pd.DataFrame) -> List[pd.DataFrame]:
    """
    Split tracks forest into trees.

    Parameters
    ----------
    tracks_df : pd.DataFrame
        DataFrame containing track information with columns:
            "track_id" : Unique identifier for each track.
            "parent_track_id" : Identifier of the parent track in the forest.
            (Other columns may be present in the DataFrame but are not used in this function.)

    Returns
    -------
    List[pd.DataFrame]
        List of dataframes, each representing a tree.
    """
    graph = tracks_df_forest(tracks_df, numba_dict=False)
    roots = graph.pop(NO_PARENT)

    tracks_by_id = tracks_df.groupby("track_id")

    trees = []
    for root in roots:
        subtree_ids = get_subtree(graph, root)
        # much faster than using pd.concat
        if len(subtree_ids) == 1:
            trees.append(tracks_by_id.get_group(root))
        else:
            trees.append(pd.concat([tracks_by_id.get_group(i) for i in subtree_ids]))

    return trees


def split_tracks_df_by_lineage(
    tracks_df: pd.DataFrame,
) -> List[pd.DataFrame]:
    """
    Split tracks dataframe into a list of dataframes, one for each lineage, sorted by the root track id.

    Parameters
    ----------
    tracks_df : pd.DataFrame
        Tracks dataframe with columns:
            "track_id" : Unique identifier for each track.
            "parent_track_id" : Identifier of the parent track in the forest.
            (Other columns may be present in the DataFrame but are not used in this function.)

    Returns
    -------
    List[pd.DataFrame]
        List of dataframes, one for each lineage.
    """
    roots = tracks_df[tracks_df["parent_track_id"] == NO_PARENT]["track_id"].to_numpy()
    lineages = [get_subgraph(tracks_df, root) for root in np.sort(roots)]
    return lineages


def get_paths_to_roots(
    tracks_df: pd.DataFrame,
    graph: Optional[Dict[int, int]] = None,
    *,
    node_index: Optional[int] = None,
    track_index: Optional[int] = None,
) -> pd.DataFrame:
    """
    Returns paths from `node_index` or `track_index` to roots.
    If `node_index` and `track_index` are None, returns all paths to roots.

    Parameters
    ----------
    tracks_df : pd.DataFrame
        DataFrame containing track information with columns:
            "track_id" : Unique identifier for each track.
            "parent_track_id" : Identifier of the parent track in the forest.
            (Other columns may be present in the DataFrame but are not used in this function.)
    graph : Optional[Dict[int, int]], optional
        Inverted forest graph, if not provided it will be computed from `tracks_df`.
    node_index : Optional[int], optional
        Node (dataframe) index to compute path to roots.
    track_index : Optional[int], optional
        Track index (track_id column value) to compute path to roots.

    Returns
    -------
    pd.DataFrame
        DataFrame containing paths to roots.
    """
    if node_index is not None and track_index is not None:
        raise ValueError("Only one of `node_index` and `track_index` can be specified.")

    if graph is None:
        graph = inv_tracks_df_forest(tracks_df)

    if node_index is not None:
        track_indices = [tracks_df.loc[node_index, "track_id"]]

    elif track_index is not None:
        track_indices = [track_index]

    else:  # both are None, return all paths
        LOG.info(
            "Both `node_index` and `track_index` are None. Returning all paths to roots."
        )
        track_indices = tracks_df["track_id"].unique().astype(int)
        parent_ids = np.asarray(list(graph.values()), dtype=int)
        track_indices = track_indices[np.isin(track_indices, parent_ids, invert=True)]

    df_by_track_id = tracks_df.groupby("track_id")

    dfs = []
    for track_id in track_indices:

        current_dfs = []
        idx = track_id
        while idx != NO_PARENT:
            current_dfs.append(df_by_track_id.get_group(idx))
            idx = graph.get(idx, NO_PARENT)

        # reverse order so that the root is the first row
        path_df = pd.concat(reversed(current_dfs), axis=0)

        # making it a single track
        path_df["track_id"] = track_id
        if "parent_track_id" in path_df.columns:
            path_df["parent_track_id"] = NO_PARENT

        dfs.append(path_df)

    out_df = pd.concat(dfs, axis=0)

    # remove nodes after node_index
    if node_index is not None:
        index = out_df.index.get_loc(node_index)
        out_df = path_df.iloc[: index + 1]

    return out_df


def _get_children(
    track_id: int,
    graph: Dict[int, List[int]],
) -> List[int]:
    """
    Returns children of `track_id` in `graph`.

    NOTE: numba graph was not working with graph.get(track_id, []) so we use this function instead.
    """
    if track_id not in graph:
        return []
    else:
        return graph[track_id]


def _filter_short_tracks(
    parent_track_id: int,
    min_length: int,
    child2parent: Dict[int, int],
    parent2children: Dict[int, List[int]],
    track_dict: Dict[int, pd.DataFrame],
    segm_painter: Optional[SegmentationPainter] = None,
) -> None:
    """
    Recursive function to short tracks created from fake division tracks shorter than `min_length`.
    All values are updated inplace.

    Parameters
    ----------
    parent_track_id : int
        Parent track id.
    min_length : int
        Minimum track length.
    child2parent : Dict[int, int]
        Child to parent track id mapping.
    parent2children : Dict[int, List[int]]
        Parent to children track id mapping.
    track_dict : Dict[int, pd.DataFrame]
        Dictionary of track dataframes.
    segm_painter : Optional[SegmentationPainter]
        Segmentation painter to update tracks segmentation label.
    """
    children_id = _get_children(parent_track_id, parent2children)

    if len(children_id) != 0 and len(children_id) != 2:
        raise ValueError(
            f"Track {parent_track_id} has {len(children_id)} children. "
            "Only tracks with 0 or 2 children are supported."
        )

    # bottom up, first fix children
    is_short_and_childless = []
    for child_id in children_id:
        # might update children length
        _filter_short_tracks(
            child_id,
            min_length,
            child2parent,
            parent2children,
            track_dict,
            segm_painter,
        )
        # check which children could be removed
        length = len(track_dict[child_id])
        is_short_and_childless.append(
            length <= min_length and len(_get_children(child_id, parent2children)) == 0
        )

    # if there's no child or all pass the filtering criteria we do not remove them.
    if sum(is_short_and_childless) == 1:

        for child_id, to_remove in zip(children_id, is_short_and_childless):
            child_track = track_dict.pop(child_id)

            if to_remove:
                new_child_track_id = 0
            else:
                # mergin sibling track
                new_child_track_id = parent_track_id
                child_track["track_id"] = new_child_track_id
                child_track["parent_track_id"] = child2parent.get(
                    parent_track_id, NO_PARENT
                )  # parent of parent track id
                track_dict[parent_track_id] = pd.concat(
                    (track_dict[parent_track_id], child_track)
                )

                # fix sibling tracks children
                child_children_id = _get_children(child_id, parent2children)

                for child_child_id in child_children_id:
                    child_child_track = track_dict[child_child_id]
                    child_child_track["parent_track_id"] = parent_track_id
                    child2parent[child_child_id] = parent_track_id

                parent2children[parent_track_id] = child_children_id

            # relabeling
            if segm_painter is not None:
                for t in child_track["t"]:
                    segm_painter.add_relabel(t, child_id, new_child_track_id)


def filter_short_sibling_tracks(
    tracks_df: pd.DataFrame,
    min_length: int,
    segments: Optional[ArrayLike] = None,
    segments_store_or_path: StoreLike = None,
    overwrite: bool = False,
) -> Union[pd.DataFrame, Tuple[pd.DataFrame, ArrayLike]]:
    """
    Filter short tracks created from fake division tracks shorter than `min_length`.

    This function tranverse the graph bottom up and remove tracks that are shorter than `min_length`
    upon divions, merging the remaining sibling track with their parent.

    If both are shorter than `min_length`, they are not removed.

    Parameters
    ----------

    tracks_df : pd.DataFrame
        DataFrame containing track information with columns:
            "track_id" : Unique identifier for each track.
            "parent_track_id" : Identifier of the parent track in the forest.
            (Other columns may be present in the DataFrame but are not used in this function.)
    min_length : int
        Minimum track length, below this value the track is removed.
    segments : Optional[ArrayLike]
        Segmentation array to update the tracks.
    segments_store_or_path : Union[Store, Path, str, None]
        Store or path to save the new segments.
    overwrite : bool
        If True, overwrite the existing segmentation array.

    Returns
    -------
    Union[pd.DataFrame, Tuple[pd.DataFrame, ArrayLike]]
        If `segments` is None, returns the modified tracks dataframe.
        If `segments` is provided, returns the modified tracks dataframe and the updated segments.
    """
    if segments is None:
        out_segments = None
        segm_painter = None
    else:
        out_segments = copy_segments(segments, segments_store_or_path, overwrite)
        segm_painter = SegmentationPainter(out_segments)

    parent2children = tracks_df_forest(tracks_df)
    child2parent = inv_tracks_df_forest(tracks_df)

    roots = parent2children.pop(NO_PARENT)

    track_dict = {
        track_id: group.copy()
        for track_id, group in tracks_df.groupby("track_id", as_index=False)
    }

    for root in roots:
        _filter_short_tracks(
            root,
            min_length,
            child2parent,
            parent2children,
            track_dict,
            segm_painter,
        )

    out_df = pd.concat(track_dict.values(), axis=0)

    if out_segments is None:
        return out_df

    segm_painter.apply_changes()

    return out_df, out_segments



import uuid
from pathlib import Path
from typing import Union

import ome_zarr
from fastapi import HTTPException
from ome_zarr.io import parse_url
from ome_zarr.reader import Reader


def create_temp_dir(path: Path, prefix: str = "", ext: str = "") -> Path:
    """Create a temporary directory.

    Parameters
    ----------
    path : Path
        The path to the directory where the temporary dir will be created.
    prefix : str
        The prefix of the temporary directory name. Defaults to "".
    ext : str
        The extension of the temporary directory name. Not required, but it is useful to
        identify zarr arrays. Defaults to "".

    Raises
    ------
    ValueError
        If the temporary directory could not be created.

    Returns
    -------
    Path
        The path to the created temporary directory
    """
    temp_dir = path / f"{prefix}{uuid.uuid4().hex}.{ext}"
    temp_dir.mkdir(exist_ok=True, parents=True)
    if not temp_dir.is_dir():
        raise ValueError(f"Error creating temporary directory {temp_dir}.")
    return temp_dir


def create_temp_zarr_path(*args, **kwargs) -> Path:
    """Create a temporary zarr directory.

    It is a wrapper around `create_temp_dir` with `ext="zarr"`.

    See Also
    --------
    create_temp_dir: original wrapped function.
    """
    return create_temp_dir(*args, ext="zarr", **kwargs)


def open_image(path: Union[str, Path]) -> ome_zarr.reader.Node:
    """Open an ome-zarr image.

    Parameters
    ----------
    path : Union[str, Path]
        The path to the ome-zarr image.

    Raises
    ------
    HTTPException
        If the image could not be opened.

    ValueError
        If the image is not a single image.

    Returns
    -------
    ome_zarr.reader.Node
        The opened image.
    """
    try:
        reader = Reader(parse_url(path))
        nodes = list(reader())
        if len(nodes) != 1:
            raise ValueError(f"Expected a single image, found {len(nodes)}")
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Could not open {path}. {str(e)}")
    return nodes[0]



import logging
from abc import abstractmethod
from typing import Any, Callable, Dict, Optional

from magicgui.widgets import Container, Label
from pydantic import BaseModel

LOG = logging.getLogger(__name__)


class BaseConfigWidget(Container):
    def __init__(self, config: BaseModel, label: Optional[str] = None):
        super().__init__()

        if label is not None:
            self.append(Label(label=label))

        self._attr_to_widget: Dict[str, Container] = {}
        self._setup_widgets()
        self.config = config

        self.native.layout().addStretch(0)

    @abstractmethod
    def _setup_widgets(self) -> None:
        pass

    @property
    def config(self) -> BaseModel:
        return self._config

    @config.setter
    def config(self, value: BaseModel) -> None:
        """Sets config and updates the sub widgets values"""
        self._config = value
        for k, v in self._config:
            # some parameters might not be exposed in the UI
            if k in self._attr_to_widget:
                widget = self._attr_to_widget[k]
                widget.changed.disconnect()
                widget.value = v
                widget.changed.connect(self._set_config_func(k))

    def _set_config_func(self, key: str) -> Callable[[Any], None]:
        """Updates config attribute and logs it"""

        def _set_config(value: Any) -> None:
            LOG.info(f"Updating {type(self).__name__} {key} with value {value}")
            setattr(self._config, key, value)

        return _set_config



import enum
import logging
import pickle
from pathlib import Path
from typing import Any, List, Optional, Union

import numpy as np
import pandas as pd
import sqlalchemy as sqla
from numpy.typing import ArrayLike
from sqlalchemy import (
    BigInteger,
    Boolean,
    Column,
    Enum,
    Float,
    ForeignKey,
    Integer,
    PickleType,
    func,
)
from sqlalchemy.engine import make_url
from sqlalchemy.orm import Session, declarative_base

from ultrack.config.dataconfig import DatabaseChoices, DataConfig
from ultrack.utils.array import assert_same_length
from ultrack.utils.constants import NO_PARENT

Base = declarative_base()

LOG = logging.getLogger(__name__)


class MaybePickleType(PickleType):
    """Only (un)pickle if value is not bytes."""

    cache_ok = True

    def bind_processor(self, dialect):
        processor = super().bind_processor(dialect)

        def _process(value):
            if isinstance(value, (bytes, memoryview)):
                return value
            try:
                return processor(value)
            except pickle.UnpicklingError:
                # for some reason, when converting database it has a few extra bytes
                return processor(bytes.fromhex(value[3:].decode("utf-8")))

        return _process

    def result_processor(self, dialect, coltype):
        processor = super().result_processor(dialect, coltype)

        def _process(value):
            if not isinstance(value, (bytes, memoryview)):
                return value
            try:
                return processor(value)
            except pickle.UnpicklingError:
                # for some reason, when converting database it has a few extra bytes
                return processor(bytes.fromhex(value[3:].decode("utf-8")))

        return _process


class NodeSegmAnnotation(enum.IntEnum):
    UNKNOWN = 0
    CORRECT = 1
    UNDERSEGMENTED = 2
    OVERSEGMENTED = 3


class VarAnnotation(enum.IntEnum):
    UNKNOWN = 0
    REAL = 1
    FAKE = 2


class NodeDB(Base):
    __tablename__ = "nodes"
    t = Column(Integer, primary_key=True)
    id = Column(BigInteger, primary_key=True, unique=True)
    parent_id = Column(BigInteger, default=NO_PARENT)
    # hierarchy parent id matches to `id` column.
    hier_parent_id = Column(BigInteger, default=NO_PARENT)
    t_node_id = Column(Integer)
    t_hier_id = Column(Integer)
    z = Column(Float)
    y = Column(Float)
    x = Column(Float)
    z_shift = Column(Float, default=0.0)
    y_shift = Column(Float, default=0.0)
    x_shift = Column(Float, default=0.0)
    area = Column(Integer)
    frontier = Column(Float, default=-1.0)
    height = Column(Float, default=-1.0)
    selected = Column(Boolean, default=False)
    pickle = Column(MaybePickleType)
    features = Column(MaybePickleType, default=None)
    node_prob = Column(Float, default=-1.0)
    segm_annot = Column(Enum(NodeSegmAnnotation), default=NodeSegmAnnotation.UNKNOWN)
    node_annot = Column(Enum(VarAnnotation), default=VarAnnotation.UNKNOWN)
    appear_annot = Column(Enum(VarAnnotation), default=VarAnnotation.UNKNOWN)
    disappear_annot = Column(Enum(VarAnnotation), default=VarAnnotation.UNKNOWN)
    division_annot = Column(Enum(VarAnnotation), default=VarAnnotation.UNKNOWN)


class OverlapDB(Base):
    __tablename__ = "overlaps"
    id = Column(Integer, primary_key=True, autoincrement=True)
    node_id = Column(BigInteger, ForeignKey(f"{NodeDB.__tablename__}.id"))
    ancestor_id = Column(BigInteger, ForeignKey(f"{NodeDB.__tablename__}.id"))


class LinkDB(Base):
    __tablename__ = "links"
    id = Column(Integer, primary_key=True, autoincrement=True)
    # source_id is the node that is before (t-1) the target_id node (t)
    source_id = Column(BigInteger, ForeignKey(f"{NodeDB.__tablename__}.id"))
    target_id = Column(BigInteger, ForeignKey(f"{NodeDB.__tablename__}.id"))
    weight = Column(Float)
    annotation = Column(Enum(VarAnnotation), default=VarAnnotation.UNKNOWN)


class GTNodeDB(Base):
    __tablename__ = "gt_nodes"
    id = Column(Integer, primary_key=True, autoincrement=True)
    t = Column(Integer)
    label = Column(Integer)
    pickle = Column(MaybePickleType)
    z = Column(Float)
    y = Column(Float)
    x = Column(Float)


class GTLinkDB(Base):
    """
    Ground-truth links database.
    These links are between GTNodes and (non-GT) Nodes.
    """

    __tablename__ = "gt_links"
    id = Column(Integer, primary_key=True, autoincrement=True)
    source_id = Column(BigInteger, ForeignKey(f"{NodeDB.__tablename__}.id"))
    target_id = Column(BigInteger, ForeignKey(f"{GTNodeDB.__tablename__}.id"))
    weight = Column(Float)
    selected = Column(Boolean, default=False)


def maximum_time_from_database(data_config: DataConfig) -> int:
    """Returns the maximum `t` found in the `NodesDB`."""
    engine = sqla.create_engine(data_config.database_path)
    with Session(engine) as session:
        max_t = session.query(func.max(NodeDB.t)).scalar()

    LOG.info(f"Found max time = {max_t}")
    if max_t is None:
        raise ValueError(f"Dataset at {data_config.database_path} is empty.")

    return max_t


def is_table_empty(data_config: DataConfig, table: Base) -> bool:
    """Checks if table is empty."""
    url = make_url(data_config.database_path)
    if (
        data_config.database == DatabaseChoices.sqlite.value
        and not Path(url.database).exists()
    ):
        # avoids creating a database with create_engine call
        return True

    engine = sqla.create_engine(data_config.database_path)
    with Session(engine) as session:
        is_empty = (
            not sqla.inspect(engine).has_table(table.__tablename__)
            or session.query(table).first() is None
        )
    return is_empty


def set_node_values(
    data_config: DataConfig,
    indices: ArrayLike,
    **kwargs,
) -> None:
    """Set arbitrary values to a node in the database given its `node_id`.

    Parameters
    ----------
    data_config : DataConfig
        Data configuration parameters.
    indices : ArrayLike
        Nodes' indices database index.
    **kwargs : Any
        Arbitrary keyword arguments to be set.
    """

    keys = list(kwargs.keys())
    kwargs["node_id"] = indices

    if isinstance(indices, int):
        for k, v in kwargs.items():
            # it might be a numpy scalar
            try:
                v = v.item()
            except AttributeError:
                pass
            kwargs[k] = [v]

    else:
        for k, v in kwargs.items():
            if hasattr(v, "tolist"):
                kwargs[k] = v.tolist()

    assert_same_length(**kwargs)

    records = [
        {k: v[i] for k, v in kwargs.items()} for i in range(len(kwargs["node_id"]))
    ]

    engine = sqla.create_engine(data_config.database_path)
    with Session(engine) as session:
        stmt = (
            sqla.update(NodeDB)
            .where(NodeDB.id == sqla.bindparam("node_id"))
            .values({k: sqla.bindparam(k) for k in keys})
        )
        session.connection().execute(
            stmt,
            records,
            execution_options={"synchronize_session": False},
        )
        session.commit()


def get_node_values(
    data_config: DataConfig,
    indices: Optional[Union[int, ArrayLike]],
    values: Union[Column, List[Column]],
) -> List[Any]:
    """Get the annotation of `node_id`.

    Parameters
    ----------
    data_config : DataConfig
        Data configuration parameters.
    indices : int
        Node database indices.
    values : List[Column]
        List of columns to be queried.
    """
    if not isinstance(values, List):
        values = [values]

    values.insert(0, NodeDB.id)

    is_scalar = False
    if isinstance(indices, int):
        indices = [indices]
        is_scalar = True

    elif isinstance(indices, np.ndarray):
        indices = indices.astype(int).tolist()

    engine = sqla.create_engine(data_config.database_path)
    with Session(engine) as session:
        query = session.query(*values)

        if indices is not None:
            query = query.where(NodeDB.id.in_(indices))

        df = pd.read_sql_query(query.statement, session.bind, index_col="id")

    if indices is not None and len(df) != len(indices):
        raise ValueError(
            f"Query returned {len(df)} rows, expected {len(indices)}."
            "\nCheck if node_id exists in database."
        )

    df = df.squeeze()
    if is_scalar:
        try:
            df = df.item()
        except ValueError:
            pass

    return df


def clear_all_data(database_path: str) -> None:
    """Clears all data from database"""
    LOG.info("Clearing all databases.")
    engine = sqla.create_engine(database_path)
    Base.metadata.drop_all(engine)



from pathlib import Path
from typing import Optional, Sequence

import click
from napari.plugins import _initialize_plugins
from napari.viewer import ViewerModel

from ultrack import link
from ultrack.cli.utils import (
    batch_index_option,
    channel_axis_option,
    config_option,
    napari_reader_option,
    overwrite_option,
    paths_argument,
)
from ultrack.config import MainConfig


@click.command("link")
@paths_argument()
@napari_reader_option()
@config_option()
@channel_axis_option(
    default=None,
    help="Channel axis, only used when input `paths` are provided",
)
@batch_index_option()
@overwrite_option()
def link_cli(
    paths: Sequence[Path],
    reader_plugin: str,
    config: MainConfig,
    channel_axis: Optional[int],
    batch_index: Optional[int],
    overwrite: bool,
) -> None:
    """Links segmentation candidates adjacent in time."""

    images = tuple()
    if len(paths) > 0:
        _initialize_plugins()

        viewer = ViewerModel()

        kwargs = {}
        if channel_axis is not None:
            kwargs["channel_axis"] = channel_axis

        images = [
            layer.data[0] if layer.multiscale else layer.data
            for layer in viewer.open(paths, **kwargs, plugin=reader_plugin)
        ]
        del viewer

    link(
        config,
        images=images,
        scale=config.data_config.metadata.get("scale"),
        batch_index=batch_index,
        overwrite=overwrite,
    )



import logging
import webbrowser
from contextlib import redirect_stderr, redirect_stdout
from typing import Any, Generator, Optional

import napari
import qtawesome as qta
import toml
from magicgui.widgets import create_widget
from napari.layers import Image, Layer
from napari.qt.threading import thread_worker
from pydantic import ValidationError
from qtpy.QtCore import Qt
from qtpy.QtGui import QCursor, QFont
from qtpy.QtWidgets import (
    QApplication,
    QComboBox,
    QFileDialog,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QSpacerItem,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from ultrack import MainConfig, export_tracks_by_extension
from ultrack.widgets.ultrackwidget.components.button_workflow_config import (
    ButtonWorkflowConfig,
)
from ultrack.widgets.ultrackwidget.components.emitting_stream import EmittingStream
from ultrack.widgets.ultrackwidget.data_forms import DataForms
from ultrack.widgets.ultrackwidget.utils import UltrackInput
from ultrack.widgets.ultrackwidget.workflows import (
    UltrackWorkflow,
    WorkflowChoice,
    WorkflowStage,
)

LOG = logging.getLogger(__name__)


def _create_link_buttons() -> list[QPushButton]:
    """
    Create buttons for external links.

    Returns
    -------
    List of QPushButton
        List of buttons with icons and link actions.
    """
    data = {
        ("Github", "mdi.github", "https://github.com/royerlab/ultrack"),
        ("BioArXiv", "mdi.file-document", ""),
        ("Documentation", "mdi.book-open", "https://royerlab.github.io/ultrack/"),
    }

    bts = []
    for name, icon_name, link in data:
        bt = QPushButton(name)
        icon = qta.icon(icon_name)
        bt.setIcon(icon)
        bt.clicked.connect(lambda _, _link=link: webbrowser.open(_link))
        bts.append(bt)

    return bts


class UltrackWidget(QWidget):
    def __init__(self, viewer: napari.Viewer) -> None:
        """
        Initialize the UltrackWidgetV2.

        Parameters
        ----------
        viewer : napari.Viewer
            The napari viewer instance.
        """
        super().__init__()
        self.viewer = viewer
        self.workflow = UltrackWorkflow(viewer)
        self._cb_images = {}
        self._data_forms = DataForms(self._on_change_config)

        self._config_valid = True
        self._all_images_are_valid = False

        self._init_ui()
        self._setup_signals()

        # Set the default workflow
        self._cb_workflow.setCurrentIndex(0)
        self._cb_workflow.currentIndexChanged.emit(0)

        self._on_image_changed(-1)
        self._on_layers_change(None)

        self._current_worker = None

    def _init_ui(self) -> None:
        """
        Initialize the user interface components.
        """
        layout = QVBoxLayout()
        self.setLayout(layout)

        self._add_title(layout)
        self._add_link_buttons(layout)
        self._add_spacer(layout, 10)
        self._add_configuration_group(layout)
        self._add_validation_messages_group(layout)
        self._add_run_button(layout)
        self._add_output_area(layout)
        self._add_cancel_button(layout)
        self._add_bt_export_tracks(layout)

        layout.addStretch()

    def _add_validation_messages_group(self, layout: QVBoxLayout) -> None:
        """
        Add the validation messages group to the layout.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the validation messages group will be added.
        """
        self._validation = QGroupBox("Validation messages")
        self._validation.setStyleSheet("QGroupBox { font-weight: bold; color: red;}")
        self._validation_messages = QLabel()
        self._validation_messages.setWordWrap(True)
        self._validation.hide()
        self._validation_messages.setStyleSheet("font-weight: normal; color: red")
        self._validation.setLayout(QVBoxLayout())
        self._validation.layout().addWidget(self._validation_messages)
        layout.addWidget(self._validation)

    def _add_bt_export_tracks(self, layout: QVBoxLayout) -> None:
        """
        Add the export tracks button to the layout.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the export tracks button will be added.
        """
        self._add_spacer(layout, 10)
        self._bt_export = QPushButton("Export tracks")
        self._bt_export.setEnabled(False)
        layout.addWidget(self._bt_export)

    def export_tracks(self):
        file_dialog = QFileDialog()
        file_dialog.setFileMode(QFileDialog.AnyFile)
        file_dialog.setAcceptMode(QFileDialog.AcceptSave)
        file_dialog.setNameFilter(
            "Napari Tracks (*.csv);;"
            "Trackmate (*.xml);;"
            "Zarr segments (*.zarr);;"
            "NetworkX (*.dot);;"
            "NetworkX (*.json)"
        )

        if file_dialog.exec_():
            file_name = file_dialog.selectedFiles()[0]
            ext = file_dialog.selectedNameFilter().split("*.")[-1][:-1]

            # add the extension if not present
            if not file_name.endswith(ext):
                file_name += f".{ext}"

            config = self._data_forms.get_config()
            export_tracks_by_extension(config, file_name, overwrite=True)

            QMessageBox.information(
                self,
                "Export tracks",
                f"Tracks exported to {file_name}",
                QMessageBox.Ok,
            )

    def _add_title(self, layout: QVBoxLayout) -> None:
        """
        Add the title and subtitle to the layout.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the title and subtitle will be added.
        """
        layout.addWidget(QLabel("<h1>ultrack</h1>"))
        subtitle = QLabel(
            "<h4>Large-scale cell tracking under segmentation uncertainty</h4>"
        )
        subtitle.setWordWrap(True)
        layout.addWidget(subtitle)

    def _add_link_buttons(self, layout: QVBoxLayout) -> None:
        """
        Add link buttons to the layout.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the link buttons will be added.
        """
        layout_bt_links = QHBoxLayout()
        for bt in _create_link_buttons():
            layout_bt_links.addWidget(bt)
        layout.addLayout(layout_bt_links)

    def _add_spacer(self, layout: QVBoxLayout, size: int = 10) -> None:
        """
        Add a spacer item to the layout.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the spacer will be added.
        size : int
            The size of the spacer.
        """
        layout.addSpacerItem(QSpacerItem(0, size))

    def _add_configuration_group(self, layout: QVBoxLayout) -> None:
        """
        Add the configuration group to the layout.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the configuration group will be added.
        """
        gp_ultrack = QGroupBox("Ultrack Configuration")
        font = QFont()
        font.setBold(True)
        gp_ultrack.setFont(font)
        gp_layout = QVBoxLayout()
        gp_ultrack.setLayout(gp_layout)

        # Description
        gp_layout.addSpacerItem(QSpacerItem(0, 5))
        self._add_configuration_description(gp_layout)

        # Workflow selector
        gp_layout.addSpacerItem(QSpacerItem(0, 10))
        self._add_workflow_selector(gp_layout)

        # Image selectors
        self._add_image_selectors(gp_layout)
        gp_layout.addSpacerItem(QSpacerItem(0, 10))

        # Settings buttons
        self._add_settings_buttons(gp_layout)

        # Form tabs
        self._add_form_tabs(gp_layout)

        layout.addSpacerItem(QSpacerItem(0, 5))
        layout.addWidget(gp_ultrack)

        self.main_group = gp_ultrack

    def _add_configuration_description(self, layout: QVBoxLayout) -> None:
        """
        Add the configuration description to the layout.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the configuration description will be added.
        """
        label = QLabel(
            "Configure the Ultrack parameters and run the tracking algorithm."
        )
        label.setWordWrap(True)
        layout.addWidget(label)

    def _add_workflow_selector(self, layout: QVBoxLayout) -> None:
        """
        Add the workflow selector to the layout.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the workflow selector will be added.
        """
        layout.addWidget(QLabel("<h5>Select the desired Ultrack workflow</h5>"))
        self._cb_workflow = QComboBox()
        for workflow_choice in WorkflowChoice:
            name = workflow_choice.value
            self._cb_workflow.addItem(name, workflow_choice)
        layout.addWidget(self._cb_workflow)

    def _add_image_selectors(self, layout: QVBoxLayout) -> None:
        """
        Add image selectors to the layout.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the image selectors will be added.
        """
        self.images = ["image", "contours", "detection", "labels"]

        for image in UltrackInput:
            image_desc = image.value
            label = QLabel(image_desc)
            layout.addWidget(label)
            widget = create_widget(annotation=Image, label=image_desc)
            widget._lb = label
            widget.native.setPlaceholderText("Select image")
            self._cb_images[image] = widget
            layout.addWidget(widget.native)

    def _add_settings_buttons(self, layout: QVBoxLayout) -> None:
        """
        Add settings buttons to the layout.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the settings buttons will be added.
        """
        button_layout = QHBoxLayout()
        self._bt_toggle_settings = self._create_toggle_settings_button()
        button_layout.addWidget(self._bt_toggle_settings)

        self._bt_load_settings = self._create_button(
            "mdi.folder-open", "Load settings from file"
        )
        button_layout.addWidget(self._bt_load_settings)

        self._bt_save_settings = self._create_button(
            "mdi.content-save", "Save settings to file"
        )
        button_layout.addWidget(self._bt_save_settings)

        layout.addLayout(button_layout)

    def _create_toggle_settings_button(self) -> QPushButton:
        """
        Create the toggle settings button.

        Returns
        -------
        QPushButton
            The created toggle settings button.
        """
        button = QPushButton("Show advanced Settings")
        button.setCheckable(True)
        button.setChecked(False)
        button.setIcon(qta.icon("mdi.chevron-down"))
        button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        return button

    def _create_button(self, icon_name: str, tooltip: str) -> QPushButton:
        """
        Create a button with an icon and tooltip.

        Parameters
        ----------
        icon_name : str
            The name of the icon.
        tooltip : str
            The tooltip text.

        Returns
        -------
        QPushButton
            The created button.
        """
        button = QPushButton(qta.icon(icon_name), "")
        button.setToolTip(tooltip)
        return button

    def _add_form_tabs(self, layout: QVBoxLayout) -> None:
        """
        Add form tabs to the layout.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the form tabs will be added.
        """
        self._tab_scroll_area = QScrollArea()
        self._tab_scroll_area.setWidgetResizable(True)
        self._tab_scroll_area.setWidget(self._data_forms.get_tab_widget())
        self._tab_scroll_area.hide()
        layout.addWidget(self._tab_scroll_area)

    def _add_run_button(self, layout: QVBoxLayout) -> None:
        """
        Add the run button to the layout.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the run button will be added.
        """
        self._bt_run = QPushButton(self)
        self._bt_run.setText("Run")
        self._bt_run.setEnabled(False)
        self._bt_run.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        button_layout = QHBoxLayout()
        button_layout.addWidget(self._bt_run)

        self._bt_run_config = ButtonWorkflowConfig(self, self.workflow)
        button_layout.addWidget(self._bt_run_config)

        layout.addLayout(button_layout)

    def _setup_signals(self) -> None:
        """Set up the signals for various buttons and widgets."""
        self._bt_toggle_settings.clicked.connect(self._on_toggle_settings)
        self._bt_save_settings.clicked.connect(self._on_save_settings)
        self._bt_load_settings.clicked.connect(self._on_load_settings)
        self._bt_export.clicked.connect(self.export_tracks)
        self._bt_run.clicked.connect(self._on_run)
        self._bt_cancel.clicked.connect(self._cancel)
        self._cb_workflow.currentIndexChanged.connect(self._on_workflow_changed)
        self.viewer.layers.events.removed.connect(self._on_layers_change)
        self.viewer.layers.events.inserted.connect(self._on_layers_change)

        for cb_image in self._cb_images.values():
            cb_image.changed.connect(self._on_image_changed)

    def _on_change_config(self):
        """Handle the change of the configuration."""
        if hasattr(self, "_bt_run"):
            try:
                new_config = self._data_forms.get_config()
                additional_config = self._data_forms.get_additional_options()
                inputs = {k: w.value for k, w in self._cb_images.items()}
                workflow = self.workflow.get_stage(
                    new_config, additional_options=additional_config, inputs=inputs
                )
                self._bt_run_config.set_workflow_stage(workflow)
                self._bt_run.setEnabled(self._all_images_are_valid)
                self._validation.hide()
                self._config_valid = True
            except ValidationError as e:
                msg = "<ul>"
                for err in e.errors():
                    msg += f"<li>{err['msg']}.</li>"
                msg += "</ul>"
                self._validation_messages.setText(msg)
                self._validation.show()
                self._bt_run.setEnabled(False)
                self._config_valid = False

    def _on_run(self):
        """Handle the run button click event.

        It runs the selected workflow with the provided inputs.
        """
        config = self._data_forms.get_config()
        additional_config = self._data_forms.get_additional_options()
        workflow_choice = self._cb_workflow.currentData()

        inputs: dict[UltrackInput, Layer] = {}

        for k, w in self._cb_images.items():
            if w.value is not None:
                inputs[k] = w.value

        worker = self._make_run_worker(
            config, workflow_choice, inputs, additional_options=additional_config
        )
        self._current_worker = worker
        worker.started.connect(self._on_run_started)
        worker.yielded.connect(self._on_workflow_intermediate_results)
        worker.finished.connect(self._on_run_finished)
        worker.start()

    def _on_workflow_intermediate_results(self, layer: Layer):
        """Handle the intermediate results of the workflow."""
        self.viewer.add_layer(layer)

    def _on_run_started(self):
        """Handle the start of the run worker."""
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        self._bt_run.setEnabled(False)
        self._bt_export.setEnabled(False)
        self._bt_run.setText("Running...")
        self._bt_run.repaint()
        self.ctx_stdout_switcher.__enter__()
        self.ctx_stderr_switcher.__enter__()
        self._output.show()
        self._bt_cancel.show()
        self._bt_cancel.setEnabled(True)
        self.main_group.setEnabled(False)

    def _on_run_finished(self):
        """Handle the finish of the run worker."""
        self._bt_run.setEnabled(True)
        self._bt_run.setText("Run")
        self._bt_export.setEnabled(
            self.workflow.last_reached_stage == WorkflowStage.DONE
        )
        self._bt_run.repaint()
        self.ctx_stdout_switcher.__exit__(None, None, None)
        self.ctx_stderr_switcher.__exit__(None, None, None)
        self._output.hide()
        self._bt_cancel.hide()
        self.main_group.setEnabled(True)
        self._current_worker = None
        QApplication.restoreOverrideCursor()

    @thread_worker
    def _make_run_worker(
        self,
        config: MainConfig,
        workflow_choice: WorkflowChoice,
        inputs: dict[UltrackInput, Layer],
        additional_options: dict[Any],
    ) -> Generator:
        """
        Create a worker to run the selected workflow.

        Returns
        -------
        Worker
            The worker to run the selected workflow.
        """
        yield from self.workflow.run(
            config, workflow_choice, inputs, additional_options
        )

    def _on_save_settings(self) -> None:
        """Handle the save settings button click event."""
        options = QFileDialog.Options()
        file_name, _ = QFileDialog.getSaveFileName(
            None,
            "Save TOML",
            "ultrack_settings.toml",
            "TOML Files (*.toml);;All Files (*)",
            options=options,
        )
        if file_name:
            with open(file_name, "w") as f:
                data = self._data_forms.get_config().model_dump(by_alias=True)
                toml.dump(data, f)
            print(f"Data saved to {file_name}")

    def _on_load_settings(self) -> None:
        """
        Handle the load settings button click event.
        """
        options = QFileDialog.Options()
        file_name, _ = QFileDialog.getOpenFileName(
            None, "Open TOML", "", "TOML Files (*.toml);;All Files (*)", options=options
        )
        if file_name:
            with open(file_name) as f:
                data = toml.load(f)
                data = MainConfig(**data)
                self._data_forms.load_config(data)

    def _on_toggle_settings(self):
        """
        Handle the toggle settings button click event.
        """
        if self._bt_toggle_settings.isChecked():
            self._tab_scroll_area.show()
            icon = qta.icon("mdi.chevron-up")
            self._bt_toggle_settings.setIcon(icon)
            self._bt_toggle_settings.setText("Hide advanced Settings")
        else:
            self._tab_scroll_area.hide()
            icon = qta.icon("mdi.chevron-down")
            self._bt_toggle_settings.setIcon(icon)
            self._bt_toggle_settings.setText("Show advanced Settings")

    def _on_layers_change(self, _: Any) -> None:
        """
        Update layer choices when layers are changed.
        """
        for widget in self._cb_images.values():
            widget.choices = self.viewer.layers

    def _on_image_changed(self, _: int) -> None:
        """
        Handle image selection changes.

        Parameters
        ----------
        _ : int
            Ignored.
        """
        self._all_images_are_valid = True

        for ultrack_input, widget in self._cb_images.items():
            if widget.enabled:
                if widget.value is None:
                    self._all_images_are_valid = False
                if ultrack_input == UltrackInput.IMAGE:
                    self._data_forms.notify_image_update(widget.value)

        self._bt_run.setEnabled(self._all_images_are_valid and self._config_valid)

    def _on_workflow_changed(self, index: int) -> None:
        """
        Handle workflow selection changes.

        Parameters
        ----------
        index : int
            The index of the selected workflow.
        """
        workflow_choice = self._cb_workflow.itemData(index)
        inputs = self.workflow.inputs_from_choice(workflow_choice)
        for name, widget in self._cb_images.items():
            if name in inputs:
                widget.native.show()
                widget.native.setEnabled(True)
                widget._lb.show()
            else:
                widget.native.hide()
                widget.native.setEnabled(False)
                widget._lb.hide()

        self._data_forms.setup_additional_options(workflow_choice)

    def _add_output_area(self, layout: QVBoxLayout) -> None:
        """
        Add the output area to the layout.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the output area will be added.
        """
        self._output = QTextEdit()
        self._output.setReadOnly(True)
        self._output.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self._output.hide()
        self.ctx_stdout_switcher = redirect_stdout(EmittingStream(self._output, "gray"))
        self.ctx_stderr_switcher = redirect_stderr(EmittingStream(self._output, "red"))
        layout.addWidget(self._output)

    def _add_cancel_button(self, layout: QVBoxLayout) -> None:
        """
        Add the cancel button to the layout.

        Parameters
        ----------
        layout : QVBoxLayout
            The layout to which the cancel button will be added.
        """
        self._bt_cancel = QPushButton(self)
        self._bt_cancel.hide()
        self._bt_cancel.setText("Cancel")
        self._bt_cancel.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        layout.addWidget(self._bt_cancel)

    def _cancel(self):
        """Cancel the current worker."""
        if self._current_worker is not None:
            self._current_worker.quit()
            self._bt_cancel.setEnabled(False)

    @staticmethod
    def find_ultrack_widget(viewer: napari.Viewer) -> Optional["UltrackWidget"]:
        """Find and returns Ultrack widget. If widget not found, returns None"""
        for _, w in viewer.window._dock_widgets.items():
            if isinstance(w.widget(), UltrackWidget):
                return w.widget()

        return None


if __name__ == "__main__":
    from napari import Viewer

    viewer = Viewer()
    widget = UltrackWidget(viewer)
    widget.show()
    viewer.window.add_dock_widget(widget, area="right")
    napari.run()



from pathlib import Path
from typing import Optional, Sequence

import click
from napari.plugins import _initialize_plugins
from napari.viewer import ViewerModel

from ultrack.cli.utils import (
    channel_axis_option,
    config_option,
    napari_reader_option,
    paths_argument,
)
from ultrack.config import MainConfig
from ultrack.imgproc.flow import add_flow


@click.command("add_flow")
@paths_argument()
@napari_reader_option()
@config_option()
@channel_axis_option(default=0, help="Coordinates shift axis.")
def add_flow_cli(
    paths: Sequence[Path],
    reader_plugin: str,
    config: MainConfig,
    channel_axis: Optional[int],
) -> None:
    """Adds coordinates shift (vector field) to segmentation hypotheses."""
    _initialize_plugins()

    viewer = ViewerModel()

    vector_field = [
        layer.data[0] if layer.multiscale else layer.data
        for layer in viewer.open(paths, channel_axis=channel_axis, plugin=reader_plugin)
    ]
    del viewer

    add_flow(
        config,
        vector_field=vector_field,
    )



from ultrack.tracks.gap_closing import close_tracks_gaps
from ultrack.tracks.graph import (
    add_track_ids_to_tracks_df,
    filter_short_sibling_tracks,
    get_paths_to_roots,
    get_subgraph,
    inv_tracks_df_forest,
    left_first_search,
    split_tracks_df_by_lineage,
    split_trees,
    tracks_df_forest,
)
from ultrack.tracks.sorting import (
    sort_track_ids,
    sort_trees_by_length,
    sort_trees_by_max_radius,
)
from ultrack.tracks.stats import (
    tracks_df_movement,
    tracks_length,
    tracks_profile_matrix,
)



import logging
from enum import Enum
from typing import Dict, Generator, List, Optional, Sequence

import napari
import numpy as np
import pandas as pd
from napari.layers import Image, Labels, Layer, Tracks
from napari.types import ArrayLike

from ultrack import MainConfig, link, segment, solve, to_tracks_layer, tracks_to_zarr
from ultrack.imgproc import detect_foreground, robust_invert
from ultrack.imgproc.flow import (
    add_flow,
    advenct_from_quasi_random,
    timelapse_flow,
    trajectories_to_tracks,
)
from ultrack.tracks import tracks_df_movement
from ultrack.utils import labels_to_contours
from ultrack.utils.array import array_apply, create_zarr
from ultrack.utils.cuda import on_gpu
from ultrack.widgets.ultrackwidget.utils import UltrackInput

LOGGER = logging.getLogger(__name__)


class WorkflowChoice(Enum):
    AUTO_DETECT = "Auto detect foreground and contours from image"
    AUTO_FROM_LABELS = "Auto detect contours from labels"
    MANUAL = "Manual selection of inputs"


class WorkflowStage(Enum):
    """
    Enumeration representing different stages of the workflow.
    """

    PREPROCESSING = "Preprocessing"
    SEGMENTATION = "Segmentation"
    LINKING = "Linking"
    SOLVING = "Tracking"
    DONE = "Done"

    def next(self):
        """
        Get the next stage in the workflow.

        Returns
        -------
        WorkflowStage
            The next stage.
        """
        members = list(WorkflowStage)
        idx = members.index(self)
        idx = min(idx + 1, len(members) - 1)
        return members[idx]


class UltrackWorkflow:
    """
    A class representing the Ultrack workflow, managing the various stages
    from preprocessing to tracking.

    Attributes
    ----------
    viewer : napari.Viewer
        The napari viewer instance.
    config : Optional[MainConfig]
        The main configuration for the workflow.
    """

    def __init__(self, viewer: napari.Viewer) -> None:
        """
        Initialize the workflow with the given viewer.

        Parameters
        ----------
        viewer : napari.Viewer
            The napari viewer instance.
        """
        self.config: Optional[MainConfig] = None
        self.additional_options = {}
        self.viewer = viewer
        self.last_reached_stage = None
        self.inputs_used = {}

    def get_stage(
        self,
        config: MainConfig,
        additional_options: dict,
        inputs: dict[UltrackInput, Layer],
    ) -> WorkflowStage:
        """
        Determine the current stage of the workflow based on the configuration.

        Parameters
        ----------
        config : MainConfig
            The main configuration for the workflow.
        additional_options : Dict
            Additional options for preprocessing.
        inputs : dict[UltrackInput, Layer]
            The layers used in the workflow.

        Returns
        -------
        WorkflowStage
            The current stage of the workflow.
        """
        if (
            self.config is None
            or self.config != config
            or self.additional_options != additional_options
            or self.last_reached_stage == WorkflowStage.PREPROCESSING
            or self.inputs_used != inputs
        ):
            return WorkflowStage.PREPROCESSING
        elif (
            self.config.segmentation_config != config.segmentation_config
            or self.last_reached_stage == WorkflowStage.SEGMENTATION
        ):
            return WorkflowStage.SEGMENTATION
        elif (
            self.config.linking_config != config.linking_config
            or self.last_reached_stage == WorkflowStage.LINKING
        ):
            return WorkflowStage.LINKING
        elif (
            self.config.tracking_config != config.tracking_config
            or self.last_reached_stage == WorkflowStage.SOLVING
        ):
            return WorkflowStage.SOLVING
        else:
            return WorkflowStage.DONE

    def notify_input_modification(self) -> None:
        """
        Notify that the input has been modified, resetting the configuration.
        """
        self.config = None
        self.additional_options = {}
        self.last_reached_stage = None
        self.inputs_used = {}

    def run(
        self,
        config: MainConfig,
        workflow_choice: WorkflowChoice,
        inputs: Dict[UltrackInput, Layer],
        additional_options: Dict,
    ) -> Generator:
        """
        Run the workflow based on the provided configuration and workflow choice.

        Parameters
        ----------
        config : MainConfig
            The main configuration for the workflow.
        workflow_choice : str
            The workflow choice indicating the preprocessing method.
        additional_options : Dict
            Additional options for preprocessing.
        inputs: Dict[UltrackInput, Layer]
            The inputs for the workflow.
        """
        image = inputs.get(UltrackInput.IMAGE, None)
        labels = inputs.get(UltrackInput.LABELS, None)
        detection = inputs.get(UltrackInput.DETECTION, None)
        contours = inputs.get(UltrackInput.CONTOURS, None)

        stage = self.get_stage(config, additional_options, inputs)
        self.inputs_used = inputs
        self.config = config
        self.additional_options = additional_options
        self.last_reached_stage = WorkflowStage.PREPROCESSING
        try:
            if stage == WorkflowStage.PREPROCESSING:
                if workflow_choice != WorkflowChoice.MANUAL:
                    if image is None and labels is None:
                        raise ValueError(
                            "Image or labels must be provided in auto mode"
                        )

                    if image is not None:
                        scale = image.scale
                    else:
                        scale = labels.scale

                    detection, contours = self._run_preprocessing(
                        image=image.data if image is not None else None,
                        labels=labels.data if labels is not None else None,
                        choice=workflow_choice,
                        **additional_options,
                    )

                    detection = Image(
                        data=detection, name="Foreground", visible=False, scale=scale
                    )
                    contours = Image(
                        data=contours, name="Contours", visible=False, scale=scale
                    )
                    yield detection
                    yield contours
                else:
                    if detection is None or contours is None:
                        raise ValueError(
                            "Foreground and contours must be provided in manual mode"
                        )
                stage = stage.next()
        except Exception as e:
            self.last_reached_stage = WorkflowStage.PREPROCESSING
            raise e

        try:
            if stage == WorkflowStage.SEGMENTATION:
                self._run_segmentation(detection, contours)
                tracks_layer, flow_field = self._add_flow(
                    image, workflow_choice, additional_options
                )
                if tracks_layer is not None:
                    yield tracks_layer
                if flow_field is not None:
                    add_flow(self.config, flow_field)

                stage = stage.next()

        except Exception as e:
            self.last_reached_stage = WorkflowStage.SEGMENTATION
            raise e

        try:
            if stage == WorkflowStage.LINKING:
                self._run_linking(detection.scale)
                stage = stage.next()
        except Exception as e:
            self.last_reached_stage = WorkflowStage.LINKING
            raise e

        try:
            if stage == WorkflowStage.SOLVING:
                self._run_solving()
                tracks_df, graph = to_tracks_layer(self.config)

                segments = tracks_to_zarr(
                    config,
                    tracks_df,
                    store_or_path=None,
                    overwrite=True,
                )

                yield Labels(
                    data=segments,
                    name="Segments",
                    visible=True,
                    scale=detection.scale,
                )

                is_3d = segments.ndim == 4
                dims = ["z", "y", "x"] if is_3d else ["y", "x"]

                yield Tracks(
                    data=tracks_df[["track_id", "t", *dims]].values,
                    graph=graph,
                    name="Tracks",
                    scale=detection.scale[1:],
                )
        except Exception as e:
            self.last_reached_stage = WorkflowStage.SOLVING
            raise e
        self.last_reached_stage = WorkflowStage.DONE
        return

    def _run_preprocessing(
        self,
        image: Optional[ArrayLike],
        labels: Optional[ArrayLike],
        choice: WorkflowChoice,
        detect_foreground_kwargs: Optional[Dict] = None,
        robust_invert_kwargs: Optional[Dict] = None,
        label_to_contours_kwargs: Optional[Dict] = None,
        **_,
    ) -> tuple[ArrayLike, ArrayLike]:
        """
        Run the preprocessing step based on the chosen method.

        Parameters
        ----------
        image : Optional[ArrayLike]
            The input image for preprocessing.
        labels : Optional[ArrayLike]
            The input labels for preprocessing.
        choice : str
            The preprocessing choice.
        detect_foreground_kwargs : Dict
            Keyword arguments for detecting foreground.
        robust_invert_kwargs : Dict
            Keyword arguments for robust inversion.
        label_to_contours_kwargs : Dict
            Keyword arguments for converting labels to contours.

        Returns
        -------
        Tuple[ArrayLike, ArrayLike]
            The detection and contours arrays.
        """
        if choice == WorkflowChoice.AUTO_DETECT:
            if detect_foreground_kwargs is None:
                LOGGER.info("No detect_foreground_kwargs provided, using default.")
                detect_foreground_kwargs = {}
            if robust_invert_kwargs is None:
                LOGGER.info("No robust_invert_kwargs provided, using default.")
                robust_invert_kwargs = {}
            return self._auto_detect(
                image, detect_foreground_kwargs, robust_invert_kwargs
            )
        elif choice == WorkflowChoice.AUTO_FROM_LABELS:
            if label_to_contours_kwargs is None:
                LOGGER.info("No label_to_contours_kwargs provided, using default.")
                label_to_contours_kwargs = {}
            return self._auto_from_labels(labels, label_to_contours_kwargs)
        else:
            raise ValueError(f"Unknown choice {choice}")

    def _add_flow(
        self, image: Layer, workflow_choice: WorkflowChoice, additional_options: dict
    ) -> tuple[Optional[Tracks], Optional[ArrayLike]]:
        """Add flow vectors to ultrack if provided.

        Parameters
        ----------
        image : ArrayLike
            The input image.
        workflow_choice : str
            The workflow choice. Either "manual" or "auto_detect".
        additional_options : dict
            Additional options for the workflow.

        Returns
        -------
        Optional[Tracks]
            The flow vectors as tracks.
        Optional[ArrayLike]
            The flow field.
        """
        if (
            workflow_choice == WorkflowChoice.MANUAL
            or workflow_choice == WorkflowChoice.AUTO_DETECT
        ):
            if additional_options.get("flow_kwargs") is not None:
                flow_kwargs = additional_options["flow_kwargs"]
                if flow_kwargs["__enable__"]:
                    del flow_kwargs["__enable__"]
                    flow_field = timelapse_flow(
                        image.data,
                        store_or_path=None,
                        **flow_kwargs,
                    )
                    flow_kwargs["__enable__"] = True

                    shape = list(image.data.shape)
                    if flow_kwargs["channel_axis"]:
                        shape.pop(flow_kwargs["channel_axis"] + 1)

                    ndim = flow_field.shape[1]
                    dd = ["dz", "dy", "dx"] if ndim == 3 else ["dy", "dx"]
                    d = ["z", "y", "x"] if ndim == 3 else ["y", "x"]
                    trajectory = advenct_from_quasi_random(
                        flow_field, shape[-ndim:], n_samples=1000
                    )

                    flow_tracklets = pd.DataFrame(
                        trajectories_to_tracks(trajectory),
                        columns=["track_id", "t", *d],
                    )
                    flow_tracklets[d] += 0.5
                    #                       ^^^ napari was crashing otherwise,
                    #                           might be an openGL issue
                    flow_tracklets[dd] = tracks_df_movement(
                        flow_tracklets, cols=tuple(d)
                    )
                    flow_tracklets["angles"] = np.arctan2(
                        # [dy, dx] or [dz, dy, dx]
                        *[flow_tracklets[d] for d in dd]
                    )

                    tracks = Tracks(
                        flow_tracklets[["track_id", "t", *d]],
                        name="flow vectors",
                        visible=False,
                        tail_length=25,
                        features=flow_tracklets[["angles", "dy", "dx"]],
                        colormap="hsv",
                    )
                    tracks.color_by = "angles"
                    return tracks, flow_field
        return None, None

    def _auto_detect(
        self,
        image: ArrayLike,
        detect_foreground_kwargs: Dict,
        robust_invert_kwargs: Dict,
    ) -> tuple[ArrayLike, ArrayLike]:
        """
        Automatically detect foreground and contours from the input image.

        Parameters
        ----------
        image : ArrayLike
            The input image.
        detect_foreground_kwargs : Dict
            Keyword arguments for detecting foreground.
        robust_invert_kwargs : Dict
            Keyword arguments for robust inversion.

        Returns
        -------
        Tuple[ArrayLike, ArrayLike]
            The detection and contours arrays.
        """
        if (
            detect_foreground_kwargs["channel_axis"]
            != robust_invert_kwargs["channel_axis"]
        ):
            raise ValueError(
                "Channel axis must be the same for both detect_foreground "
                "and robust_invert."
            )

        channel = detect_foreground_kwargs["channel_axis"]

        shape = list(image.shape)
        if channel:
            shape.pop(channel + 1)

        detection, contours = _create_temp_detection_and_contours(shape)

        array_apply(
            image,
            out_array=contours,
            func=on_gpu(robust_invert),
            **robust_invert_kwargs,
        )

        array_apply(
            image,
            out_array=detection,
            func=on_gpu(detect_foreground),
            **detect_foreground_kwargs,
        )

        return detection, contours

    def _auto_from_labels(
        self, labels: ArrayLike, label_to_contours_kwargs: Dict
    ) -> None:
        """
        Automatically create contours from input labels.

        Parameters
        ----------
        labels : ArrayLike
            The input labels.
        label_to_contours_kwargs : Dict
            Keyword arguments for converting labels to contours.

        Returns
        -------
        Tuple[ArrayLike, ArrayLike]
            The detection and contours arrays.
        """
        zarr_detection, zarr_contours = labels_to_contours(
            labels,
            **label_to_contours_kwargs,
        )

        return zarr_detection, zarr_contours

    def _run_segmentation(self, detection: Layer, contours: Layer) -> None:
        """
        Run the segmentation step.

        Parameters
        ----------
        detection : napari.layers.Layer
            The detection napari layer.
        contours : napari.layers.Layer
            The contours napari layer.
        """
        segment(
            detection=detection.data,
            edge=contours.data,
            config=self.config,
            overwrite=True,
        )

    def _run_linking(self, scale: Sequence[float]) -> None:
        """
        Run the linking step.

        Parameters
        ----------
        scale : Sequence[float]
            The scale of the input data.
        """
        link(config=self.config, overwrite=True, scale=scale)

    def _run_solving(self) -> None:
        """
        Run the solving step.
        """
        solve(config=self.config, overwrite=True)

    def inputs_from_choice(self, choice: WorkflowChoice) -> List[UltrackInput]:
        """
        Get the required inputs based on the workflow choice.

        Parameters
        ----------
        choice : str
            The workflow choice.

        Returns
        -------
        List[str]
            A list of required inputs.
        """
        if choice == WorkflowChoice.AUTO_DETECT:
            return [UltrackInput.IMAGE]
        elif choice == WorkflowChoice.AUTO_FROM_LABELS:
            return [UltrackInput.LABELS]
        elif choice == WorkflowChoice.MANUAL:
            return [UltrackInput.DETECTION, UltrackInput.CONTOURS]
        else:
            raise ValueError(f"Unknown choice {choice}")


def _create_temp_detection_and_contours(
    shape: tuple[int, ...],
) -> tuple[ArrayLike, ArrayLike]:
    """
    Create temporary storage for detection and contours arrays.

    Parameters
    ----------
    shape : Tuple[int, ...]
        The shape of the arrays.

    Returns
    -------
    Tuple[ArrayLike, ArrayLike]
        The paths to the temporary detection and contours arrays.
    """
    zarr_detection = create_zarr(
        shape=shape,
        dtype=bool,
        overwrite=True,
    )
    zarr_contours = create_zarr(
        shape=shape,
        dtype=np.float16,
        overwrite=True,
    )

    return zarr_detection, zarr_contours



from typing import Tuple

import numpy as np
import pytest
import zarr

from ultrack.utils.estimation import estimate_parameters_from_labels


def test_timelapse_parameter_estimation(
    timelapse_mock_data: Tuple[zarr.Array, zarr.Array, zarr.Array],
) -> None:
    _, _, labels = timelapse_mock_data
    df = estimate_parameters_from_labels(labels, is_timelapse=True)
    assert not np.any(df["area"] == np.nan)
    # there's zero drift between time points on mock data
    assert np.all(df["distance"] == 0.0)


@pytest.mark.parametrize(
    "segmentation_mock_data",
    [{"n_dim": 2}, {"n_dim": 3}],
    indirect=True,
)
def test_stack_parameter_estimation(
    segmentation_mock_data: Tuple[np.ndarray, np.ndarray, np.ndarray],
) -> None:
    _, _, labels = segmentation_mock_data
    df = estimate_parameters_from_labels(labels, is_timelapse=False)
    assert not np.any(df["area"] == np.nan)



from pathlib import Path
from typing import Any, Dict, List, Tuple

import numpy as np
import pytest
import toml
import zarr

from ultrack.config.config import MainConfig, load_config
from ultrack.config.dataconfig import DatabaseChoices
from ultrack.core.linking.processing import link
from ultrack.core.segmentation.processing import segment
from ultrack.core.solve.processing import solve
from ultrack.utils.data import (
    make_cell_division_mock_data,
    make_config_content,
    make_segmentation_mock_data,
)

# from testing.postgresql import Postgresql


@pytest.fixture
def config_content(tmp_path: Path, request) -> Dict[str, Any]:
    kwargs = {"data.working_dir": str(tmp_path)}
    if hasattr(request, "param"):
        kwargs.update(request.param)

    # FIXME: needs to fork testing.postgresql
    # if postgresql create dummy server and close when done
    is_postgresql = kwargs.get("data.database") == DatabaseChoices.postgresql.value

    if is_postgresql:
        # FIXME: not working, falling back to sqlite
        kwargs["data.database"] = DatabaseChoices.sqlite.value

    return make_config_content(kwargs)


@pytest.fixture
def config_path(tmp_path: Path, config_content: Dict[str, Any]) -> Path:
    path = tmp_path / "config.toml"
    with open(path, mode="w") as f:
        toml.dump(config_content, f)
    return path


@pytest.fixture
def config_instance(config_path: Path) -> MainConfig:
    return load_config(config_path)


@pytest.fixture
def zarr_dataset_paths(
    tmp_path: Path, timelapse_mock_data: Tuple[zarr.Array, zarr.Array]
) -> List[str]:

    paths = []
    for src_array, filename in zip(
        timelapse_mock_data, ("foreground.zarr", "contours.zarr", "labels.zarr")
    ):
        path = tmp_path / filename
        out_store = zarr.storage.LocalStore(path)
        zarr.save(out_store, src_array[...])
        paths.append(str(path))

    return paths


@pytest.fixture
def segmentation_mock_data(request) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    return make_segmentation_mock_data(**request.param)


@pytest.fixture
def timelapse_mock_data(request) -> Tuple[zarr.Array, zarr.Array, zarr.Array]:
    if not hasattr(request, "param"):
        request.param = {}

    # avoiding popping from instance
    kwargs = request.param.copy()
    length = kwargs.pop("length", 3)

    blobs, contours, labels = make_segmentation_mock_data(**kwargs)
    shape = (length,) + blobs.shape

    foreground = zarr.empty(
        shape,
        store=zarr.storage.MemoryStore(),
        chunks=(1, *blobs.shape),
        dtype=blobs.dtype,
    )
    edge = zarr.empty(
        shape,
        store=zarr.storage.MemoryStore(),
        chunks=(1, *blobs.shape),
        dtype=contours.dtype,
    )
    segmentation = zarr.empty(
        shape,
        store=zarr.storage.MemoryStore(),
        chunks=(1, *blobs.shape),
        dtype=labels.dtype,
    )

    for t in range(length):
        foreground[t] = blobs
        edge[t] = contours
        segmentation[t] = labels

    return foreground, edge, segmentation


@pytest.fixture
def segmentation_database_mock_data(
    config_instance: MainConfig,
    timelapse_mock_data: Tuple[zarr.Array, zarr.Array, zarr.Array],
) -> MainConfig:
    foreground, edge, _ = timelapse_mock_data
    segment(
        foreground,
        edge,
        config_instance,
    )
    return config_instance


@pytest.fixture(scope="function")
def linked_database_mock_data(
    segmentation_database_mock_data: MainConfig,
) -> MainConfig:
    config = segmentation_database_mock_data
    link(config)
    return config


@pytest.fixture(scope="function")
def tracked_database_mock_data(
    linked_database_mock_data: MainConfig,
) -> MainConfig:
    config = linked_database_mock_data
    solve(config)
    return config


@pytest.fixture
def cell_division_mock_data() -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    return make_cell_division_mock_data()


@pytest.fixture(scope="function")
def tracked_cell_division_mock_data(
    cell_division_mock_data: Tuple[np.ndarray, np.ndarray, np.ndarray],
    config_instance: MainConfig,
) -> MainConfig:
    foreground, contours, _ = cell_division_mock_data

    segment(
        foreground,
        contours,
        config_instance,
    )
    link(config_instance)
    solve(config_instance)

    return config_instance



import logging
from typing import List, Optional

import numpy as np
import pandas as pd

from ultrack.tracks.sorting import sort_track_ids
from ultrack.utils.constants import NO_PARENT

LOG = logging.getLogger(__name__)


def tracks_df_movement(
    tracks_df: pd.DataFrame,
    lag: int = 1,
    cols: Optional[tuple[str, ...]] = None,
) -> pd.DataFrame:
    """
    Compute the displacement for track data across given time lags.

    This function computes the displacement (difference) for track coordinates
    across the specified lag periods.

    NOTE: this sort the dataframe by ["track_id", "t"].

    Parameters
    ----------
    tracks_df : pd.DataFrame
        Dataframe containing track data. It is expected to have columns
        ["track_id", "t"] and any of ["z", "y", "x"] representing the 3D coordinates.

    lag : int, optional
        Number of periods to compute the difference over. Default is 1.

    cols : tuple[str, ...], optional
        Columns to compute the displacement for. If not provided, it will try to
        find any of ["z", "y", "x"] columns in the dataframe and use them.

    Returns
    -------
    pd.DataFrame
        Dataframe of the displacement (difference) of coordinates for the given lag.
        Displacements for the first row of each track_id will be set to zero.

    Examples
    --------
    >>> df = pd.DataFrame({
    ...     "track_id": [1, 1, 2, 2],
    ...     "t": [1, 2, 1, 2],
    ...     "z": [0, 1, 0, 2],
    ...     "y": [1, 2, 1, 2],
    ...     "x": [2, 3, 2, 2]
    ... })
    >>> print(tracks_df_movement(df))
       z    y    x
    0 0.0  0.0  0.0
    1 1.0  1.0  1.0
    2 0.0  0.0  0.0
    3 2.0  1.0  0.0
    """
    if "track_id" not in tracks_df.columns:
        tracks_df["track_id"] = 0

    tracks_df.sort_values(by=["track_id", "t"], inplace=True)

    if cols is None:
        cols = []
        for c in ["z", "y", "x"]:
            if c in tracks_df.columns:
                cols.append(c)
    else:
        cols = list(cols)

    out = tracks_df.groupby("track_id", as_index=False)[cols].diff(periods=lag)
    out.fillna(0, inplace=True)

    return out


def estimate_drift(df: pd.DataFrame, quantile: float = 0.99) -> float:
    """Compute a estimate of the tracks drift.

    NOTE: this sort the dataframe by ["track_id", "t"].

    Parameters
    ----------
    df : pd.DataFrame
        Tracks dataframe, must have `track_id` column.
    quantile : float, optional
        Drift quantile, by default 0.99

    Returns
    -------
    float
        Drift from the given quantile.
    """
    displacement = tracks_df_movement(df)
    assert all(c in ["z", "y", "x"] for c in displacement.columns)
    distances = np.linalg.norm(displacement, axis=1)
    robust_max_distance = np.quantile(distances, quantile)
    LOG.info(f"{quantile} quantile spatial drift distance of {robust_max_distance}")
    return robust_max_distance


def tracks_profile_matrix(
    tracks_df: pd.DataFrame,
    columns: List[str],
) -> np.ndarray:
    """
    Construct a profile matrix from a pandas DataFrame containing tracks data.

    Parameters
    ----------
    tracks_df : pd.DataFrame
        DataFrame containing track information with columns:
            "track_id" : Unique identifier for each track.
            "t" : Time step index for each data point in the track.
            Other columns specified in 'columns' parameter, representing track attributes.
    columns : List[str]
        List of strings, specifying the columns of 'tracks_df' to use as attributes.

    Returns
    -------
    np.ndarray
        A 3D NumPy array representing the profile matrix with shape (num_attributes, num_tracks, max_timesteps),
        where 'num_attributes' is the number of attributes specified in 'columns',
        'num_tracks' is the number of unique tracks, and 'max_timesteps' is the maximum number of timesteps
        encountered among all tracks.
    """
    extra_cols = []
    for column in columns:
        if column not in tracks_df.columns:
            extra_cols.append(column)

    if len(extra_cols) > 0:
        raise ValueError(f"Columns {extra_cols} not in {tracks_df.columns}.")

    if "parent_track_id" in tracks_df.columns:
        sorted_track_ids = sort_track_ids(tracks_df)
    else:
        sorted_track_ids = tracks_df["track_id"].unique()

    num_t = int(tracks_df["t"].max(skipna=True) + 1)

    profile_matrix = np.zeros((len(sorted_track_ids), num_t, len(columns)), dtype=float)

    by_track_id = tracks_df.groupby("track_id")

    for i, track_id in enumerate(sorted_track_ids):
        group = by_track_id.get_group(track_id)
        t = group["t"].to_numpy(dtype=int)
        profile_matrix[np.full_like(t, i), t] = group[columns].values

    # move attribute axis to the first dimension
    profile_matrix = profile_matrix.transpose((2, 0, 1))
    try:
        profile_matrix = np.squeeze(profile_matrix, axis=0)
    except ValueError:
        pass

    return profile_matrix


def tracks_length(
    tracks_df: pd.DataFrame,
    include_appearing: bool = True,
    include_disappearing: bool = True,
) -> pd.DataFrame:
    """
    Compute the length of each track in a tracks dataframe.

    Parameters
    ----------
    tracks_df : pd.DataFrame
        DataFrame containing track information with columns:
            "t" : Time step index for each data point in the track.
            "track_id" : Unique identifier for each track.
            "parent_track_id" : Unique identifier for the parent track.
    include_appearing : bool, optional
        Include tracks that appear outside the first time step, by default True.
    include_disappearing : bool, optional
        Include tracks that disappear outside the last time step, by default True.

    Returns
    -------
    pd.DataFrame
        Series containing the length of each track.
    """
    for c in ("t", "track_id", "parent_track_id"):
        if c not in tracks_df.columns:
            raise ValueError(f"Column '{c}' not in {tracks_df.columns}.")

    tracks_df = tracks_df.groupby("track_id", as_index=False).agg(
        start=pd.NamedAgg("t", lambda x: x.min()),
        end=pd.NamedAgg("t", lambda x: x.max()),
        parent_track_id=pd.NamedAgg("parent_track_id", lambda x: x.iloc[0]),
    )
    tracks_df["length"] = tracks_df["end"] - tracks_df["start"] + 1

    t_max = tracks_df["end"].max()

    if not include_appearing:
        tracks_df = tracks_df[
            (tracks_df["start"] == 0) | (tracks_df["parent_track_id"] != NO_PARENT)
        ]

    if not include_disappearing:
        tracks_df = tracks_df[
            (tracks_df["end"] == t_max)
            | (tracks_df["track_id"].isin(tracks_df["parent_track_id"]))
        ]

    return tracks_df



import pytest
import sqlalchemy as sqla
from sqlalchemy.orm import Session

from ultrack.config.config import MainConfig
from ultrack.core.database import LinkDB, NodeDB
from ultrack.core.linking.utils import clear_linking_data
from ultrack.utils.constants import NO_PARENT


@pytest.mark.parametrize(
    "config_content",
    [
        {
            "segmentation.n_workers": 4,
            "linking.n_workers": 4,
        }
    ],
    indirect=True,
)
def test_clear_linking_data(
    tracked_database_mock_data: MainConfig,
) -> None:
    database_path = tracked_database_mock_data.data_config.database_path

    clear_linking_data(database_path)

    engine = sqla.create_engine(database_path)
    with Session(engine) as session:
        assert session.query(LinkDB).count() == 0
        assert session.query(NodeDB).where(NodeDB.selected).count() == 0
        assert session.query(NodeDB).where(NodeDB.parent_id != NO_PARENT).count() == 0
        # not checking overlap because our mock data doesn't have any overlap
        assert session.query(NodeDB).count() > 0



import logging
import os
import pickle
from contextlib import nullcontext
from typing import Callable, List, Optional

import fasteners
import numpy as np
import pandas as pd
import sqlalchemy as sqla
import zarr
from numpy.typing import ArrayLike
from skimage.measure._regionprops import RegionProperties, regionprops_table
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session
from toolz import curry

from ultrack.config.config import MainConfig, SegmentationConfig
from ultrack.core.database import (
    Base,
    NodeDB,
    OverlapDB,
    clear_all_data,
    get_node_values,
)
from ultrack.core.segmentation.hierarchy import create_hierarchies
from ultrack.core.segmentation.node import Node
from ultrack.utils.array import check_array_chunk
from ultrack.utils.deprecation import rename_argument
from ultrack.utils.multiprocessing import (
    batch_index_range,
    multiprocessing_apply,
    multiprocessing_sqlite_lock,
)

logging.basicConfig()
logging.getLogger("sqlachemy.engine").setLevel(logging.INFO)

LOG = logging.getLogger(__name__)


def _generate_id(index: int, time: int, max_segments: int) -> int:
    """Generates a unique id.

    Parameters
    ----------
    index : int
        Current index per time.
    time_point : int
        Current time.
    max_segments : int
        Upper bound of number of segments.

    Returns
    -------
    int
        Unique id.
    """
    return index + (time + 1) * max_segments


def _insert_db(
    engine: Engine, time: int, nodes: List[NodeDB], overlaps: List[OverlapDB]
) -> None:
    """
    Helper function to insert data into database.
    IMPORTANT: This function resets `nodes` and `overlaps`.

    Parameters
    ----------
    engine : Engine
        SQLAlchemy connection engine.
    time : int
        Current time.
    nodes : List[NodeDB]
        List of nodes to insert.
    overlaps : List[OverlapDB]
        List of overlap of nodes to insert.
    """
    LOG.info(f"Pushing some nodes from hier time {time} to database.")
    with Session(engine) as session:
        session.add_all(nodes)
        session.add_all(overlaps)
        session.commit()
    engine.dispose()
    nodes.clear()
    overlaps.clear()


class _ImageCachedLazyLoader:
    """
    Wrapper class to cache dask/zarr data loading for feature computation.
    """

    def __init__(self, image: ArrayLike):
        self._image = image
        self._current_t = -1
        self._frame = None

    def __getitem__(self, index: int) -> np.ndarray:
        if index != self._current_t:
            self._frame = np.asarray(self._image[index])
            self._current_t = index
        return self._frame


def create_feats_callback(
    shape: ArrayLike, image: Optional[ArrayLike], properties: List[str]
) -> Callable[[Node], np.ndarray]:
    """
    Create a callback function to compute features for each node.

    Parameters
    ----------
    shape : ArrayLike
        Volume (plane) shape.
    image : Optional[ArrayLike]
        Image array for segments properties, could have channel dimension on last axis.
    properties : List[str]
        List of properties to compute for each segment, see skimage.measure.regionprops documentation.

    Returns
    -------
    Callable[[Node], np.ndarray]
        Callback function to compute features for each node returning a numpy array.
    """
    mask = np.zeros(shape, dtype=bool)

    if image is not None:
        image = _ImageCachedLazyLoader(image)

    def _feats_callback(node: Node) -> np.ndarray:

        node.paint_buffer(mask, True, include_time=False)

        if image is None:
            frame = None
        else:
            frame = image[node.time]

        obj = RegionProperties(
            node.slice,
            label=True,
            label_image=mask,
            intensity_image=frame,
            cache_active=True,
        )

        feats = np.concatenate(
            [np.ravel(getattr(obj, p)) for p in properties], dtype=np.float32
        )

        node.paint_buffer(mask, False, include_time=False)

        return feats

    return _feats_callback


@curry
def _process(
    time: int,
    foreground: ArrayLike,
    contours: ArrayLike,
    config: SegmentationConfig,
    db_path: str,
    max_segments_per_time: int,
    write_lock: Optional[fasteners.InterProcessLock] = None,
    catch_duplicates_expection: bool = False,
    insertion_throttle_rate: int = 50,
    images: Optional[ArrayLike] = None,
    properties: Optional[List[str]] = None,
) -> None:
    """Process `foreground` and `edge` of current time and add data to database.

    Parameters
    ----------
    time : int
        Current time.
    foreground : ArrayLike
        Foreground array.
    contours : ArrayLike
        Contours array.
    config : SegmentationConfig
        Segmentation configuration parameters.
    db_path : str
        Path to database including type prefix.
    max_segments_per_time : int
        Upper bound of segments per time point.
    lock : Optional[fasteners.InterProcessLock], optional
        Lock object for SQLite multiprocessing, optional otherwise, by default None.
    catch_duplicates_expection : bool
        If True, catches duplicates exception, by default False.
    insertion_throttle_rate : int
        Throttling rate for insertion, by default 50.
    image : Optional[ArrayLike], optional
        Image array for segments properties, channel dimension is optional on last axis, by default None.
    properties : Optional[List[str]], optional
        List of properties to compute for each segment, by default None.
    """
    if config.random_seed == "frame":
        np.random.seed(time)

    edge_map = contours[time]
    if config.max_noise > 0:
        noise = np.random.uniform(0, config.max_noise, size=edge_map.shape)
        # promoting edge_map to smallest float
        noise = noise.astype(np.result_type(edge_map.dtype, np.float16))
        edge_map = edge_map + noise

    hiers = create_hierarchies(
        foreground[time] > config.threshold,
        edge_map,
        hierarchy_fun=config.ws_hierarchy,
        max_area=config.max_area,
        min_area=config.min_area,
        min_area_factor=config.min_area_factor,
        min_frontier=config.min_frontier,
    )

    LOG.info(f"Computing nodes of time {time}")

    engine = sqla.create_engine(db_path, hide_parameters=True)

    index = 1
    nodes = []
    overlaps = []

    node_feats = None
    feats_callback = None

    if properties is not None:
        feats_callback = create_feats_callback(foreground.shape[1:], images, properties)

    for h, hierarchy in enumerate(hiers):
        hierarchy.cache = True

        hier_index_map = {}
        for hier_node in hierarchy.nodes:
            hier_node.id = _generate_id(index, time, max_segments_per_time)
            hier_node.time = time
            hier_node._parent = None  # avoiding pickling parent hierarchy
            centroid = hier_node.centroid

            if len(centroid) == 2:
                y, x = centroid
                z = 0
            else:
                z, y, x = centroid

            if feats_callback is not None:
                node_feats = feats_callback(hier_node)

            node = NodeDB(
                id=hier_node.id,
                t_node_id=index,
                t_hier_id=h + 1,
                t=time,
                z=int(z),
                y=int(y),
                x=int(x),
                area=int(hier_node.area),
                frontier=hier_node.frontier,
                height=hier_node.height,
                pickle=pickle.dumps(hier_node),  # pickling to reduce memory usage
                features=node_feats,
            )

            hier_index_map[hier_node._h_node_index] = node
            nodes.append(node)

            index += 1
            del hier_node

        tree = hierarchy.tree

        for h_index, node in hier_index_map.items():

            h_parent_index = tree.parent(h_index)
            # checking if h_index is root or parent is not in hier_index_map
            if h_index != h_parent_index:
                try:
                    # assign hierarchy parent to node
                    node.hier_parent_id = hier_index_map[h_parent_index].id
                except KeyError:
                    pass

            # find overlapping segments by iterating through hierarchy (tree)
            for a in tree.ancestors(h_index):
                if a in hier_index_map and a != h_index:
                    overlaps.append(
                        OverlapDB(
                            node_id=hier_index_map[h_index].id,
                            ancestor_id=hier_index_map[a].id,
                        )
                    )

            LOG.info(
                f"{len(overlaps)} overlaps found for node {hier_index_map[h_index].id}."
            )

        hierarchy.cache = False
        hierarchy.free_props()

        if index > max_segments_per_time:
            raise ValueError(
                f"Number of segments exceeds upper bound of {max_segments_per_time} per time."
            )

        try:
            # if lock is None it inserts at every iteration
            if write_lock is None:
                if (h + 1) % insertion_throttle_rate == 0:  # throttling insertions
                    _insert_db(engine, time, nodes, overlaps)

            # otherwise it inserts only when it can acquire the lock
            elif write_lock.acquire(blocking=False):
                _insert_db(engine, time, nodes, overlaps)
                write_lock.release()

        except sqla.exc.IntegrityError as e:
            if (
                catch_duplicates_expection
                and "duplicate key value violates unique constraint" in str(e).lower()
            ):
                LOG.warning(
                    f"IntegrityError: {e}\nSkipping batch {time}.\n"
                    "Cannot guarantee image was entirely processed.\nSEGMENTS MIGHT BE MISSING!"
                )
                return
            else:
                raise ValueError(
                    "Duplicated nodes found. Set `overwrite=True` to overwrite existing data."
                ) from e

    # pushes any remaning data
    with write_lock if write_lock is not None else nullcontext():
        _insert_db(engine, time, nodes, overlaps)

    LOG.info(f"DONE with time {time}.")


def _check_zarr_memory_store(arr: ArrayLike) -> None:
    if isinstance(arr, zarr.Array) and isinstance(arr.store, zarr.storage.MemoryStore):
        LOG.warning(
            "Found zarr with MemoryStore. "
            "Using an zarr with MemoryStore can lead to considerable memory usage."
        )


def _get_properties_names(
    shape: ArrayLike,
    image: Optional[ArrayLike],
    properties: Optional[List[str]],
) -> Optional[List[str]]:
    """
    Get properties names from provided properties list.

    Parameters
    ----------
    shape : ArrayLike
        Volume (plane) shape including time.
    image : Optional[ArrayLike]
        Image array for segments properties, could have channel dimension on last axis.
    properties : Optional[List[str]]
        List of properties to compute for each segment, see skimage.measure.regionprops documentation.
    """

    if properties is None:
        return None

    ndim = len(shape) - 1

    if image is None:
        dummy_image = None
    else:
        if image.ndim == len(shape):  # no channel dimension
            dummy_image = np.ones((4,) * ndim, dtype=np.float32)
        else:
            # adding channel dimension
            dummy_image = np.ones((4,) * ndim + (image.shape[-1],), dtype=np.float32)

    dummy_labels = np.zeros((4,) * ndim, dtype=np.uint32)
    dummy_labels[:2, :2] = 1

    data_dict = regionprops_table(dummy_labels, dummy_image, properties=properties)

    return list(data_dict.keys())


@rename_argument("detection", "foreground")
@rename_argument("edge", "contours")
def segment(
    foreground: ArrayLike,
    contours: ArrayLike,
    config: MainConfig,
    max_segments_per_time: int = 1_000_000,
    batch_index: Optional[int] = None,
    overwrite: bool = False,
    insertion_throttle_rate: int = 50,
    images: Optional[ArrayLike] = None,
    properties: Optional[List[str]] = None,
) -> None:
    """Add candidate segmentation (nodes) from `foreground` and `edge` to database.

    Parameters
    ----------
    foreground : ArrayLike
        Foreground probability array of shape (T, (Z), Y, X)
    contours : ArrayLike
        Contours array of shape (T, (Z), Y, X)
    config : MainConfig
        Configuration parameters.
    max_segments_per_time : int
        Upper bound of segments per time point.
    batch_index : Optional[int], optional
        Batch index for processing a subset of nodes, by default everything is processed.
    overwrite : bool
        Cleans up segmentation, linking, and tracking database content before processing.
    insertion_throttle_rate : int
        Throttling rate for insertion, by default 50.
        Only used with non-sqlite databases.
    images : Optional[ArrayLike], optional
        Image array of shape (T, (Z), Y, X, (C)) for segments properties, by default None.
        Channel and Z dimensions are optional.
    properties : Optional[List[str]], optional
        List of properties to compute for each segment, see skimage.measure.regionprops documentation.
    """
    LOG.info(f"Adding nodes with SegmentationConfig:\n{config.segmentation_config}")

    if foreground.shape != contours.shape:
        raise ValueError(
            f"`foreground` and `contours` shape must match. Found {foreground.shape} and {contours.shape}"
        )

    check_array_chunk(foreground)
    check_array_chunk(contours)

    _check_zarr_memory_store(foreground)
    _check_zarr_memory_store(contours)

    LOG.info(f"Foreground array with shape {foreground.shape}")
    LOG.info(f"Edge array with shape {contours.shape}")

    length = foreground.shape[0]
    time_points = batch_index_range(
        length, config.segmentation_config.n_workers, batch_index
    )
    LOG.info(f"Segmenting time points {time_points}")

    batch_start_index = int(os.getenv("ULTRACK_BATCH_INDEX_START", "0"))

    if batch_index is None or batch_index == batch_start_index:
        engine = sqla.create_engine(config.data_config.database_path)

        if overwrite:
            clear_all_data(config.data_config.database_path)

        Base.metadata.create_all(engine)
        config.data_config.metadata_add(
            {
                "shape": foreground.shape,
                "properties": _get_properties_names(
                    foreground.shape, images, properties=properties
                ),
            }
        )

    with multiprocessing_sqlite_lock(config.data_config) as lock:
        process = _process(
            foreground=foreground,
            contours=contours,
            config=config.segmentation_config,
            db_path=config.data_config.database_path,
            write_lock=lock,
            max_segments_per_time=max_segments_per_time,
            catch_duplicates_expection=batch_index is not None,
            insertion_throttle_rate=insertion_throttle_rate,
            images=images,
            properties=properties,
        )

        multiprocessing_apply(
            process,
            time_points,
            config.segmentation_config.n_workers,
            desc="Adding nodes to database",
        )


def get_nodes_features(
    config: MainConfig,
    indices: Optional[ArrayLike] = None,
    include_persistence: bool = False,
) -> pd.DataFrame:
    """
    Creates a pandas dataframe from nodes features defined during segmentation
    plus area and coordinates.

    Parameters
    ----------
    config : MainConfig
        Configuration parameters.
    indices : Optional[ArrayLike], optional
        List of node indices, by default
    include_persistence : bool, optional
        Include persistence features, by default False

    Returns
    -------
    pd.DataFrame
        Dataframe with nodes features
    """
    feats_cols = [NodeDB.t, NodeDB.z, NodeDB.y, NodeDB.x, NodeDB.area, NodeDB.features]
    if include_persistence:
        feats_cols += [NodeDB.hier_parent_id, NodeDB.height]

    df: pd.DataFrame = get_node_values(
        config.data_config, indices=indices, values=feats_cols
    )

    if "features" in df.columns:
        feat_columns = config.data_config.metadata["properties"]
        feat_mat = np.asarray(df["features"].tolist())
        df.loc[:, feat_columns] = feat_mat
        df.drop(columns=["features"], inplace=True)

    if include_persistence:

        df.rename(columns={"height": "node_death"}, inplace=True)

        min_height = df["node_death"].min()
        max_height = df["node_death"].max()
        eps = 1e-5

        df.loc[(df["node_death"] < 0) & df["node_death"].isna(), "node_death"] = (
            max_height + eps
        )

        children_df = df[df["hier_parent_id"] > 0]

        df.loc[
            children_df["hier_parent_id"].to_numpy(),
            "node_birth",
        ] = children_df["node_death"].to_numpy()

        df.loc[df["node_birth"].isna(), "node_birth"] = min_height - eps

        df.drop(columns="hier_parent_id", inplace=True)

    return df



from typing import List, Optional

import numpy as np
from numpy.typing import ArrayLike


def _parse_voxel_size(
    voxel_size: Optional[ArrayLike],
    ndim: int,
    channel_axis: Optional[int],
) -> np.ndarray:
    """Parses voxel size and returns it in the correct format."""
    ndim = ndim - (channel_axis is not None)

    if voxel_size is None:
        voxel_size = (1.0,) * ndim

    if len(voxel_size) != ndim:
        raise ValueError(
            "Voxel size must have the same number of elements as the image dimensions."
            f"Expected {ndim} got {len(voxel_size)}."
        )

    return np.asarray(voxel_size)


def _channel_iterator(
    image: ArrayLike,
    channel_axis: Optional[int],
) -> List[Optional[int]]:
    """Iterates over channels, it returns a [None] if channel_axis is None."""

    if channel_axis is None:
        channel_iterator = [None]
    else:
        channel_iterator = range(image.shape[channel_axis])

    return channel_iterator



import qtawesome as qta
from qtpy.QtWidgets import QAction, QMenu, QPushButton

from ultrack.widgets.ultrackwidget.workflows import UltrackWorkflow, WorkflowStage


class ButtonWorkflowConfig(QPushButton):
    def __init__(self, parent, workflow: UltrackWorkflow):
        super().__init__(parent)
        self.setIcon(qta.icon("mdi.cog"))
        self.workflow = workflow
        self._setup_menu()
        self._setup_signals()

    def _setup_menu(self):
        self.menu = QMenu(self)
        # run all
        action = QAction("All", self.menu)
        action.setCheckable(True)
        action.setChecked(True)
        self.action_all = action
        self.menu.addAction(action)

        for workflow in WorkflowStage:
            if workflow == WorkflowStage.DONE:
                continue
            action = QAction(workflow.value, self.menu)
            action.setCheckable(True)
            action.setChecked(False)
            action.setData(workflow)
            self.menu.addAction(action)

    def _on_check_all(self):
        checked = self.action_all.isChecked()
        for action in self.menu.actions():
            action.setChecked(checked)

    def _on_check_workflow(self):
        # checked = self.sender().isChecked()
        # workflow = self.sender().data()

        self.action_all.triggered.disconnect(self._on_check_all)
        self.action_all.setChecked(False)
        self.action_all.triggered.connect(self._on_check_all)

    def _setup_signals(self):
        self.action_all.triggered.connect(self._on_check_all)
        for action in self.menu.actions():
            if action != self.action_all:
                action.triggered.connect(self._on_check_workflow)

    def set_workflow_stage(self, stage: WorkflowStage):
        self.action_all.setChecked(False)
        for action in self.menu.actions():
            if action.data() == stage:
                action.setChecked(True)
                stage = stage.next()
            else:
                action.setChecked(False)

    def get_workflow_stages(self) -> list[WorkflowStage]:
        stages = []
        for action in self.menu.actions():
            if action.isChecked():
                stages.append(action.data())
        return stages

    def mousePressEvent(self, e):
        # self._bt_run_options.mapToGlobal(event.pos)
        self.menu.exec_(self.mapToGlobal(e.pos()))
        super().mousePressEvent(e)



import platform

import numpy as np
import pytest
import sqlalchemy as sqla
from sqlalchemy.orm import Session

from ultrack.config import MainConfig
from ultrack.core.database import Base, NodeDB
from ultrack.imgproc.flow import (
    add_flow,
    advenct_field,
    timelapse_flow,
    trajectories_to_tracks,
)


@pytest.mark.skipif(
    platform.machine() == "arm64", reason="Test skipped on Apple Silicon"
)
@pytest.mark.parametrize("ndim", [2, 3])
def test_flow_field(ndim: int, request) -> None:
    th = pytest.importorskip("torch")

    size = (64,) * ndim
    sigma = 15
    im_factor = 2
    grid_factor = 4

    grid = th.stack(th.meshgrid([th.arange(s) for s in size], indexing="ij"), dim=-1)

    mus = th.Tensor(
        [[0.5, 0.5, 0.5], [0.55, 0.5, 0.5], [0.57, 0.48, 0.53], [0.55, 0.45, 0.55]]
    )[:, :ndim]

    mus = (mus * th.tensor(size)).round().int()

    frames = th.stack([th.exp(-th.square(grid - mu).sum(dim=-1) / sigma) for mu in mus])

    fields = timelapse_flow(
        frames.cpu().numpy(),
        im_factor=im_factor,
        grid_factor=grid_factor,
        num_iterations=2_000,
        lr=0.0001,
    )
    trajectory = advenct_field(fields, mus[None, 0], size)
    tracks = trajectories_to_tracks(trajectory)

    if request.config.getoption("--show-napari-viewer"):
        import napari

        kwargs = {"blending": "additive", "interpolation3d": "nearest", "rgb": False}

        viewer = napari.Viewer()

        viewer.add_image(frames.cpu().numpy(), **kwargs)
        viewer.add_tracks(tracks)
        viewer.add_image(
            fields,
            colormap="turbo",
            scale=(im_factor,) * ndim,
            channel_axis=1,
            **kwargs,
        )

        napari.run()

    # tolerance is super loose, should be improved, but algorithm is not precise
    th.testing.assert_close(trajectory.squeeze(), mus.half(), atol=0.0, rtol=0.05)


@pytest.mark.parametrize("n_channels,n_dim", [(2, 2), (3, 3), (2, 3)])
def test_add_flow(
    config_instance: MainConfig,
    n_channels: int,
    n_dim: int,
) -> None:
    # non-square shape
    whole_shape = (213, 157, 256)
    shape = whole_shape[-n_dim:]
    n_time_pts = 2

    data_config = config_instance.data_config

    engine = sqla.create_engine(data_config.database_path)
    Base.metadata.create_all(engine)
    data_config.metadata_add({"shape": (n_time_pts,) + shape})

    time = [1, 1, 1]
    coordinates = np.asarray(
        [
            [100, 100, 100],  # (0, 0, 0)
            [40, 140, 255],  # (0, 1, 1)
            [172, 56, 200],  # (1, 0, 1)
        ]
    )
    coordinates[:, :-n_dim] = 0.0

    vec_coordinates = coordinates.astype(float) / whole_shape
    vec_coordinates = np.round(vec_coordinates).astype(int)
    # vec_coordinates = [
    #     (0, 0, 0),
    #     (0, 1, 1),
    #     (1, 0, 1),
    # ]

    vector_field = np.arange(2**n_dim).reshape((2,) * n_dim)
    vector_field = np.tile(vector_field, (n_time_pts,) + (1,) * n_dim)
    vector_field = [vector_field * i for i in range(1, n_channels + 1)]
    """
    2D vector field
    y:
    [[0, 1],
     [2, 3]]
    x:
    [[0, 2],
     [4, 6]]

    3D vector field:

    z (y):
    [[0, 1], [[4, 5],
     [2, 3]]  [6, 7]]

    y (x):
    [[0, 2], [[8, 10],
     [4, 6]]  [12, 14]]

    x (optional):
    [[0, 3], [[12, 15],
     [6, 9]]  [18, 21]]
    """

    # normalizing vector field
    vector_field = [v / s for v, s in zip(vector_field, shape[-n_channels:])]

    with Session(engine) as session:
        for i, (t, centroids) in enumerate(zip(time, coordinates)):
            mock_node = NodeDB(
                id=i + 1,
                t_node_id=-1,
                t_hier_id=-1,
                area=0.0,
                t=t,
                z=centroids[0],
                y=centroids[1],
                x=centroids[2],
                pickle=b"0000",
            )
            session.add(mock_node)
        session.commit()

    add_flow(config_instance, vector_field)

    with Session(engine) as session:
        query = session.query(
            NodeDB.id,
            NodeDB.z_shift,
            NodeDB.y_shift,
            NodeDB.x_shift,
        )
        results = {i: (z, y, x) for i, z, y, x in query}

    for i, c in enumerate(vec_coordinates):
        index = (time[i],) + tuple(c[-n_dim:])
        expected_shift = tuple(
            v[index] * s for v, s in zip(vector_field, shape[-n_channels:])
        )
        expected_shift = (0,) * (3 - len(expected_shift)) + expected_shift
        assert results[i + 1] == expected_shift



import time
from pathlib import Path
from typing import Callable, List

import napari
import numpy as np
import pytest

from ultrack.config import MainConfig
from ultrack.widgets.ultrackwidget import UltrackWidget
from ultrack.widgets.ultrackwidget.utils import UltrackInput
from ultrack.widgets.ultrackwidget.workflows import WorkflowChoice


@pytest.mark.parametrize(
    "workflow",
    [
        WorkflowChoice.AUTO_DETECT,
        WorkflowChoice.MANUAL,
        WorkflowChoice.AUTO_FROM_LABELS,
    ],
)
@pytest.mark.parametrize("multichannel", [True, False])
def test_ultrack_widget(
    make_napari_viewer: Callable[[], napari.Viewer],
    zarr_dataset_paths: List[str],
    workflow: WorkflowChoice,
    tmp_path: Path,
    multichannel: bool,
) -> None:
    if multichannel and workflow != WorkflowChoice.AUTO_DETECT:
        pytest.skip("Multichannel only supported for auto-detect workflow")

    viewer = make_napari_viewer()
    widget = UltrackWidget(viewer)

    layers = viewer.open(zarr_dataset_paths)  # foreground, edge, label

    if multichannel:
        data = np.asarray(layers[0].data, dtype=float)
        image_data = np.stack([data, data, data], axis=-1)
    else:
        image_data = np.asarray(layers[0].data, dtype=float)

    viewer.add_image(image_data, name="image")

    layers = viewer.layers

    workflow_idx = widget._cb_workflow.findData(workflow)
    widget._cb_workflow.setCurrentIndex(workflow_idx)
    widget._cb_workflow.currentIndexChanged.emit(workflow_idx)

    # setting combobox choices manually, because they were not working automatically
    widget._cb_images[UltrackInput.DETECTION].choices = layers
    widget._cb_images[UltrackInput.CONTOURS].choices = layers
    widget._cb_images[UltrackInput.LABELS].choices = layers
    widget._cb_images[UltrackInput.IMAGE].choices = layers

    # selecting layers
    widget._cb_images[UltrackInput.DETECTION].value = layers[0]
    widget._cb_images[UltrackInput.CONTOURS].value = layers[1]
    widget._cb_images[UltrackInput.LABELS].value = layers[2]
    widget._cb_images[UltrackInput.IMAGE].value = layers[3]

    ####################################################################################
    # CHECK IF THE CHANNEL SELECTION IS PROPERLY PROPAGATED
    ####################################################################################
    if workflow == WorkflowChoice.AUTO_DETECT:
        additional_options = widget._data_forms.get_additional_options()
        expected_channel_value = 3 if multichannel else None
        assert (
            additional_options["robust_invert_kwargs"]["channel_axis"]
            == expected_channel_value
        )
        assert (
            additional_options["detect_foreground_kwargs"]["channel_axis"]
            == expected_channel_value
        )

    ####################################################################################
    # CHECK IF THE CONFIGURATION IS PROPERLY PROPAGATED
    ####################################################################################
    config = MainConfig()
    config.segmentation_config.threshold = 0.65

    # temporary working dir
    config.data_config.working_dir = tmp_path

    widget._data_forms.load_config(config)

    propagated_config = widget._data_forms.get_config()

    assert propagated_config.segmentation_config.threshold == 0.65

    ####################################################################################
    # CHECK IF THE WIDGET CAN RUN
    ####################################################################################
    widget._bt_run.clicked.emit()

    time.sleep(1)
    while widget._current_worker is not None and widget._current_worker.is_running:
        time.sleep(0.5)



from typing import Dict, List, Tuple

import pandas as pd

from ultrack.config.config import MainConfig
from ultrack.core.export.utils import solution_dataframe_from_sql
from ultrack.tracks.graph import add_track_ids_to_tracks_df, inv_tracks_df_forest


def to_tracks_layer(
    config: MainConfig,
    include_parents: bool = True,
    include_node_ids: bool = True,
) -> Tuple[pd.DataFrame, Dict[int, List[int]]]:
    """Exports solution from database to napari tracks layer format.

    Parameters
    ----------
    config : MainConfig
        Configuration parameters.
    include_parents : bool
        Flag to include parents track id for each track id.
    include_ids : bool
        Flag to include node ids for each unit.

    Returns
    -------
    Tuple[pd.DataFrame, Dict[int, List[int]]]
        Tracks dataframe and an lineage graph, mapping node_id -> parent_id.
    """
    df = solution_dataframe_from_sql(config.data_config.database_path)
    df = add_track_ids_to_tracks_df(df)
    df.sort_values(by=["track_id", "t"], inplace=True)

    graph = inv_tracks_df_forest(df)

    data_dim = len(config.data_config.metadata["shape"])
    if data_dim == 4:
        columns = ["track_id", "t", "z", "y", "x"]
    elif data_dim == 3:
        columns = ["track_id", "t", "y", "x"]
    else:
        raise ValueError(
            f"Expected dataset with 3 or 4 dimensions, T(Z)YX. Found {data_dim}."
        )

    if include_node_ids:
        df["id"] = df.index
        columns.append("id")

    if include_parents:
        columns.append("parent_track_id")

        if include_node_ids:
            columns.append("parent_id")

    return df[columns], graph



from pathlib import Path

import click
import toml

from ultrack.config.config import MainConfig


@click.command("create_config")
@click.argument("output_path", type=click.Path(path_type=Path), default="config.toml")
def config_cli(output_path: Path) -> None:
    """Creates a configuration file with default values."""
    config = MainConfig()

    if output_path.exists():
        raise ValueError(f"{output_path} already exists.")

    with open(output_path, mode="w") as f:
        toml.dump(config.model_dump(by_alias=True), f)



import numpy as np
import pytest

from ultrack.imgproc import sam


def test_sam() -> None:
    image = np.random.rand(100, 100)

    try:
        seg_model = sam.MicroSAM()
    except ModuleNotFoundError:
        pytest.skip("MicroSAM not installed")

    seg_model(image)

    # with maxima prompt
    seg_model = sam.set_peak_maxima_prompt(
        seg_model,
        sigma=5,
        min_distance=10,
        threshold_rel=0.1,
    )
    seg_model(image)



import numpy as np
import pytest

from ultrack.config.config import MainConfig
from ultrack.core.database import (
    LinkDB,
    NodeDB,
    OverlapDB,
    clear_all_data,
    get_node_values,
    is_table_empty,
    set_node_values,
)
from ultrack.core.segmentation.processing import _generate_id


@pytest.mark.parametrize(
    "config_content",
    [
        {
            "segmentation.n_workers": 4,
            "linking.n_workers": 4,
        }
    ],
    indirect=True,
)
def test_clear_all_data(
    tracked_database_mock_data: MainConfig,
) -> None:
    data_config = tracked_database_mock_data.data_config

    clear_all_data(data_config.database_path)

    assert is_table_empty(data_config, NodeDB)
    assert is_table_empty(data_config, OverlapDB)
    assert is_table_empty(data_config, LinkDB)


def test_set_get_node_values(
    segmentation_database_mock_data: MainConfig,
) -> None:

    # test single node
    index = _generate_id(1, 1, 1_000_000)

    set_node_values(
        segmentation_database_mock_data.data_config,
        index,
        area=0,
    )

    value = get_node_values(
        segmentation_database_mock_data.data_config,
        index,
        [NodeDB.area],
    )

    assert value == 0

    # test multiple nodes
    indices = np.asarray([_generate_id(i, 1, 1_000_000) for i in range(1, 3)])

    set_node_values(
        segmentation_database_mock_data.data_config,
        indices,
        area=list(range(len(indices))),
    )

    value = get_node_values(
        segmentation_database_mock_data.data_config,
        indices,
        NodeDB.area,
    )

    np.testing.assert_array_equal(value, np.arange(len(indices)))



import logging
from enum import Enum
from typing import Optional, Tuple

import napari
import numpy as np
import pandas as pd
from numpy.typing import ArrayLike
from qtpy.QtCore import Qt
from qtpy.QtGui import QKeySequence
from qtpy.QtWidgets import QLabel, QPushButton, QVBoxLayout, QWidget

from ultrack.utils.constants import NO_PARENT

LOG = logging.getLogger(__name__)


class Annotation(Enum):
    """
    Enum class representing possible annotation states.

    Attributes
    ----------
    UNLABELED : int
        Represents an unlabeled state.
    CORRECT : int
        Represents a correct annotation.
    INCORRECT : int
        Represents an incorrect annotation.
    SKIPPED : int
        Represents a skipped annotation.
    """

    UNLABELED = 0
    CORRECT = 1
    INCORRECT = 2
    SKIPPED = 3


class LinkValidation(QWidget):
    """
    A QWidget subclass for validating and annotating tracks.

    Parameters
    ----------
    images : ArrayLike
        The image data as an array.
    tracks_df : pd.DataFrame
        Dataframe containing track data.
        Must contain columns: "id", "parent_track_id", "parent_id", "t", "y", "x".
    viewer : napari.Viewer
        The napari viewer instance.
    crop_size : int, optional
        Size for cropping the image, by default 64.
    scale : Optional[ArrayLike], optional
        Scale of the image, by default None which means ones are used.
    parent : Optional[QWidget], optional
        Parent widget, by default None.
    """

    def __init__(
        self,
        *images: ArrayLike,
        tracks_df: pd.DataFrame,
        viewer: napari.Viewer,
        crop_size: int = 64,
        scale: Optional[ArrayLike] = None,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent)

        self._current_idx = 0
        self._rng = np.random.default_rng(1)

        self._annotation_df = pd.DataFrame(
            {"annotation": Annotation.UNLABELED}, index=_filter_non_pairs(tracks_df)
        )

        size = self._annotation_df.shape[0]
        self._ordering = self._rng.integers(size, size=(size,))

        self._tracks_df = tracks_df.set_index("id", drop=False)
        if "z" not in self._tracks_df.columns:
            self._tracks_df["z"] = 0

        self._images = list(images)

        if scale is None:
            self._scale = np.ones(self.ndim)
        else:
            self._scale = np.asarray(scale)

        self.setLayout(self._createLayout())

        # setups viewer
        self._viewer = viewer
        self._crop_size = np.broadcast_to(crop_size, shape=self.ndim)
        self._crop_size = self._crop_size * self._scale / self._scale.max()
        self._crop_size = self._crop_size.astype(int)
        self._half_crop = self._crop_size // 2

        # setups layers
        empty_image = np.zeros(self._crop_size, dtype=np.uint8)
        kwargs = dict(blending="additive", scale=self._scale)

        self._frame_layers = [
            self._viewer.add_image(empty_image, name="frame", colormap="red", **kwargs)
            for _ in range(len(self._images))
        ]
        self._prev_frame_layers = [
            self._viewer.add_image(
                empty_image, name="prev. frame", colormap="green", **kwargs
            )
            for _ in range(len(self._images))
        ]

        for i in range(len(self._images)):
            self._frame_layers[i]._keep_auto_contrast = True
            self._prev_frame_layers[i]._keep_auto_contrast = True

        self._pts_layer = self._viewer.add_points(
            scale=scale,
            ndim=self.ndim,
            name="centrois",
            out_of_slice_display=True,
            face_color="transparent",
        )

        # setups view
        self.current_idx = 0

    def _createLayout(self) -> QVBoxLayout:
        """
        Creates and returns a QVBoxLayout containing UI elements for the widget.

        Returns
        -------
        QVBoxLayout
            The created layout containing buttons and labels for the widget.
        """
        layout = QVBoxLayout()

        # Create buttons
        self.btn_correct = QPushButton("Correct", self)
        self.btn_incorrect = QPushButton("Incorrect", self)
        self.btn_skip = QPushButton("Skip", self)
        self.btn_undo = QPushButton("Undo", self)
        self.btn_done = QPushButton("Done", self)
        self.lb_counter = QLabel(self.counter_text, self)

        # Set tools tips
        self.btn_correct.setToolTip("Mark pair as correct (Shortcut: 1)")
        self.btn_incorrect.setToolTip("Mark pair as incorrect (Shortcut: 2)")
        self.btn_skip.setToolTip("Skip pair (Shortcut: 3)")
        self.btn_undo.setToolTip("Undo last annotation (Shortcut: 4)")
        self.btn_done.setToolTip("Close viewer and save annotations")

        # Set shortcuts
        self.btn_correct.setShortcut(QKeySequence(Qt.Key_1))
        self.btn_incorrect.setShortcut(QKeySequence(Qt.Key_2))
        self.btn_skip.setShortcut(QKeySequence(Qt.Key_3))
        self.btn_undo.setShortcut(QKeySequence(Qt.Key_4))

        # Add buttons to the layout
        layout.addWidget(self.btn_correct)
        layout.addWidget(self.btn_incorrect)
        layout.addWidget(self.btn_skip)
        layout.addWidget(self.btn_undo)
        layout.addWidget(self.btn_done)
        layout.addWidget(self.lb_counter)

        # Add actions
        self.btn_correct.clicked.connect(
            lambda: self._update_annotation(Annotation.CORRECT, 1)
        )
        self.btn_incorrect.clicked.connect(
            lambda: self._update_annotation(Annotation.INCORRECT, 1)
        )
        self.btn_undo.clicked.connect(
            lambda: self._update_annotation(Annotation.UNLABELED, -1)
        )
        self.btn_skip.clicked.connect(
            lambda: self._update_annotation(Annotation.SKIPPED, 1)
        )

        return layout

    @property
    def counter_text(self) -> str:
        """
        Returns the text displaying the current counter status.
        """
        i = self.current_idx + 1
        return f"Sample: {i} ({i / self._annotation_df.shape[0]:0.3f} %)"

    @property
    def ndim(self) -> int:
        return self._images[0].ndim - 1

    @property
    def current_idx(self) -> int:
        return self._current_idx

    @current_idx.setter
    def current_idx(self, value: int) -> None:
        """
        Setter for the current index. Also updates button states and view.

        Parameters
        ----------
        value : int
            New current index to set from 0 to number of samples - 1.
        """
        self.btn_correct.setEnabled(True)
        self.btn_incorrect.setEnabled(True)
        self.btn_skip.setEnabled(True)
        self.btn_undo.setEnabled(True)

        if value == 0:
            self.btn_undo.setEnabled(False)

        elif value == self._annotation_df.shape[0] - 1:
            self.btn_correct.setEnabled(False)
            self.btn_incorrect.setEnabled(False)
            self.btn_skip.setEnabled(False)

        self._current_idx = value
        LOG.info(f"Current index: {self._current_idx}")
        self.lb_counter.setText(self.counter_text)
        self._update_view()

    def _update_annotation(
        self,
        annotation: Annotation,
        increment: int,
    ) -> None:
        """
        Updates the annotation for the current index and logs the annotation.

        Parameters
        ----------
        annotation : Annotation
            The new annotation value.
        increment : int
            Amount to increment the current index by after updating the annotation.
        """
        self._annotation_df.iloc[self._ordering[self._current_idx]] = annotation
        LOG.info(f"Annotation: {annotation} to {self._current_idx}")
        if increment != 0:
            self.current_idx += increment

    def _crop_image(
        self, image: ArrayLike, t: int, coords: ArrayLike
    ) -> Tuple[ArrayLike, ArrayLike]:
        """
        Crop image around provided coordinates.

        Parameters
        ----------
        image : ArrayLike
            Image to crop.
        t : int
            Time point.
        coords : ArrayLike
            Coordinates around which to crop.

        Returns
        -------
        Tuple[ArrayLike, ArrayLike]
            Cropped image and lower bounds of the cropped region.
        """
        coords = np.asarray(coords, dtype=int)[-self.ndim :]

        # Compute crop bounds
        shape = np.asarray(image.shape[-self.ndim :])
        lower = np.clip(coords - self._half_crop, 0, None)
        upper = np.clip(coords + self._crop_size - self._half_crop, None, shape - 1)

        # Crop image
        slicing = (t,) + tuple(slice(l, u) for l, u in zip(lower, upper))
        crop = np.asarray(image[slicing])

        return crop, lower

    def _update_view(self) -> None:
        """
        Updates the viewer layers based on the current index and related track data.
        """
        node = self._tracks_df.iloc[self._ordering[self._current_idx]]
        parent_node = self._tracks_df.loc[int(node["parent_id"])]

        # Update frame layer
        for l1, l2, im in zip(
            self._frame_layers, self._prev_frame_layers, self._images
        ):
            l1.data, offset = self._crop_image(
                im, node["t"].item(), node[["z", "y", "x"]]
            )
            l2.data, _ = self._crop_image(
                im, parent_node["t"].item(), node[["z", "y", "x"]]
            )

        # Update points layer
        centorids = (
            np.stack([node[["z", "y", "x"]], parent_node[["z", "y", "x"]]])[
                :, -self.ndim :
            ]
            - offset[None, :]
        )
        self._pts_layer.data = centorids
        self._pts_layer.edge_color = np.array([[1, 0, 0], [0, 1, 0]])

        # Center view
        positions = self._scale * centorids[0, -len(self._scale) :]
        n_dims = len(self._viewer.dims.order)
        start_dim = n_dims - self.ndim

        self._viewer.dims.set_point(range(start_dim, n_dims), positions)

    @property
    def annotations(self) -> pd.DataFrame:
        return self._annotation_df

    @annotations.setter
    def annotations(self, value: pd.DataFrame) -> None:
        """
        Update annotations by mergin with existing annotations and moving to the next unlabeled sample.

        Parameters
        ----------
        value : pd.DataFrame
            Dataframe with annotations.
        """
        if "id" in value.columns:
            value = value.set_index("id", drop=False)["annotation"]

        self._annotation_df.update(value)
        labels = self._annotation_df["annotation"].to_numpy()
        annotated = labels != Annotation.UNLABELED
        annot_ids = np.where(annotated)[0]
        not_annot_ids = np.where(~annotated)[0]
        self._rng.shuffle(not_annot_ids)
        self._ordering = np.concatenate([annot_ids, not_annot_ids])
        self.current_idx = len(annot_ids)


def _filter_non_pairs(tracks_df: pd.DataFrame) -> ArrayLike:
    """Filter out divisions, appearing or disappearing tracks."""

    tracks_df = tracks_df.set_index("id", drop=False)

    # Filter out appearing cells
    tracks_df = tracks_df[tracks_df["parent_id"] != NO_PARENT]

    # Filter division events
    parents_track_ids = tracks_df["parent_track_id"]
    tracks_df = tracks_df[tracks_df["parent_track_id"] == parents_track_ids]

    return tracks_df["id"].to_numpy(dtype=int)


def validate_links(
    *images: ArrayLike,
    tracks_df: pd.DataFrame,
    annotations: Optional[pd.DataFrame] = None,
    crop_size: int = 64,
    scale: Optional[ArrayLike] = None,
    viewer: Optional[napari.Viewer] = None,
) -> pd.DataFrame:
    """
    Validates links by creating a GUI interface in napari for manual annotation.

    Parameters
    ----------
    images : ArrayLike
        Images data as an array.
    tracks_df : pd.DataFrame
        Dataframe containing track data.
    annotations : Optional[pd.DataFrame], optional
        Dataframe containing previous annotations, by default None.
    crop_size : int, optional
        Size for cropping the image, by default 64.
    scale : Optional[ArrayLike], optional
        Scale of the image, by default None which means ones are used.
    viewer : Optional[napari.Viewer], optional
        Existing napari viewer instance, by default None, in which case a new viewer is created.

    Returns
    -------
    pd.DataFrame
        Dataframe containing annotations from the validation session.
    """

    if viewer is None:
        viewer = napari.Viewer()

    widget = LinkValidation(
        *images,
        tracks_df=tracks_df,
        viewer=viewer,
        crop_size=crop_size,
        scale=scale,
    )

    widget.btn_done.clicked.connect(viewer.close)

    if annotations is not None:
        widget.annotations = annotations

    viewer.window.add_dock_widget(widget, area="right")

    napari.run()



import logging
from typing import Iterator

import numpy as np
import scipy.ndimage as ndi
from numpy.typing import ArrayLike
from skimage import measure, morphology
from skimage.measure._regionprops import RegionProperties

from ultrack.core.segmentation.node import Node
from ultrack.core.segmentation.vendored.hierarchy import Hierarchy as _Hierarchy
from ultrack.core.segmentation.vendored.hierarchy import oversegment_components

LOG = logging.getLogger(__name__)


class Hierarchy(_Hierarchy):
    def __init__(self, props: RegionProperties, **kwargs) -> None:
        super().__init__(props, cache=False, **kwargs)
        self.bbox = props.bbox

    def free_props(self) -> None:
        """Frees region properties content."""
        del self.props
        self.props = None

    @staticmethod
    def create_node(node_index: int, parent: "Hierarchy", **kwargs) -> Node:
        """Overrides function to create nodes, see parent class documentation."""
        return Node(node_index, parent=parent, **kwargs)


def create_hierarchies(
    binary_foreground: ArrayLike,
    edge: ArrayLike,
    **kwargs,
) -> Iterator[Hierarchy]:
    """Computes a collection of hierarchical watersheds inside `binary_foreground` mask.

    Parameters
    ----------
    binary_foreground : ArrayLike
        Binary array showing regions of interest.

    edge : ArrayLike
        Fuzzy contour image representing instances boundaries.

    Returns
    -------
    Iterator[Hierarchy]
        List of hierarchical watersheds.
    """
    binary_foreground = np.asarray(binary_foreground)

    assert (
        issubclass(binary_foreground.dtype.type, np.integer)
        or binary_foreground.dtype == bool
    )

    LOG.info("Labeling connected components.")
    labels, num_labels = ndi.label(binary_foreground, output=np.int32)
    del binary_foreground

    if "min_area" in kwargs and num_labels > 1:
        LOG.info("Filtering small connected components.")
        #  To avoid removing small objects, divide by 4 to remove the smallest ones.
        #  Nodes in hierarchies are still filtered by minimum area.
        #  This is mainly for lonely cells.
        morphology.remove_small_objects(
            labels,
            min_size=int(kwargs["min_area"] / kwargs.get("min_area_factor", 4.0)),
            out=labels,
        )

    edge = np.asarray(edge)

    if "max_area" in kwargs:
        LOG.info("Oversegmenting connected components.")
        labels = oversegment_components(
            labels,
            edge,
            kwargs["max_area"],
            kwargs.get("anisotropy_pen", 0.0),
        )

    kwargs.pop("min_area_factor")

    LOG.info("Creating hierarchies (lazy).")
    for c in measure.regionprops(labels, edge, cache=True):
        yield Hierarchy(c, **kwargs)
